import swapper
from baseapp_core.graphql.models import RelayModel
from baseapp_core.models import random_name_in
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone
from django.utils.timesince import timesince as djtimesince
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel


class AbstractBaseChatRoom(TimeStampedModel, RelayModel):
    title = models.CharField(max_length=255, blank=True, null=True)
    image = models.ImageField(
        _("image"), upload_to=random_name_in("chat_room_images"), blank=True, null=True
    )

    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="created_rooms", on_delete=models.CASCADE
    )
    last_message = models.ForeignKey(
        swapper.get_model_name("baseapp_chats", "Message"),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    last_message_time = models.DateTimeField(null=True, blank=True, default=timezone.now)
    participants_count = models.IntegerField(default=0)
    messages_count = models.IntegerField(default=0)
    is_group = models.BooleanField(default=False)

    class Meta:
        abstract = True
        ordering = ["-last_message_time", "-created"]

    def __str__(self):
        return str(self.pk)

    @classmethod
    def get_graphql_object_type(cls):
        from .graphql.object_types import ChatRoomObjectType

        return ChatRoomObjectType


class AbstractBaseMessage(TimeStampedModel, RelayModel):
    class Verbs(models.IntegerChoices):
        SENT_MESSAGE = 100, _("sent a message")

    class MessageType(models.IntegerChoices):
        USER_MESSAGE = 1, _("User message")
        SYSTEM_GENERATED = 2, _("System message")

    @property
    def description(self):
        return self.label

    content = models.TextField(null=True, blank=True)
    content_linked_profile_actor = models.ForeignKey(
        swapper.get_model_name("baseapp_profiles", "Profile"),
        related_name="linked_as_content_actor_set",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    content_linked_profile_target = models.ForeignKey(
        swapper.get_model_name("baseapp_profiles", "Profile"),
        related_name="linked_as_content_target_set",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True, blank=True
    )
    profile = models.ForeignKey(
        swapper.get_model_name("baseapp_profiles", "Profile"),
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )

    message_type = models.IntegerField(
        choices=MessageType.choices, default=MessageType.USER_MESSAGE
    )
    verb = models.IntegerField(choices=Verbs.choices, default=Verbs.SENT_MESSAGE, db_index=True)

    room = models.ForeignKey(
        swapper.get_model_name("baseapp_chats", "ChatRoom"),
        blank=True,
        null=True,
        related_name="messages",
        on_delete=models.CASCADE,
        db_index=True,
    )

    in_reply_to = models.ForeignKey(
        swapper.get_model_name("baseapp_chats", "Message"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="replies",
    )

    action_object_content_type = models.ForeignKey(
        ContentType,
        blank=True,
        null=True,
        related_name="action_object",
        on_delete=models.CASCADE,
        db_index=True,
    )
    action_object_object_id = models.IntegerField(blank=True, null=True, db_index=True)
    action_object = GenericForeignKey("action_object_content_type", "action_object_object_id")

    extra_data = models.JSONField(blank=True, null=True)

    class Meta:
        abstract = True
        ordering = ["-created"]

    def __str__(self):
        ctx = {
            "actor": (
                self.profile if self.message_type == self.MessageType.USER_MESSAGE else "The system"
            ),
            "verb": self.__class__.Verbs(self.verb).label,
            "action_object": self.action_object,
            "target": self.room,
            "timesince": self.timesince(),
        }
        if self.room:
            if self.action_object:
                return (
                    _("%(actor)s %(verb)s %(action_object)s on %(target)s %(timesince)s ago") % ctx
                )
            return _("%(actor)s %(verb)s on %(target)s %(timesince)s ago") % ctx
        if self.action_object:
            return _("%(actor)s %(verb)s %(action_object)s %(timesince)s ago") % ctx
        return _("%(actor)s %(verb)s %(timesince)s ago") % ctx

    def timesince(self, now=None):
        """
        Shortcut for the ``django.utils.timesince.timesince`` function of the
        current timestamp.
        """
        return (
            djtimesince(self.created, now).encode("utf8").replace(b"\xc2\xa0", b" ").decode("utf8")
        )

    @classmethod
    def get_graphql_object_type(cls):
        from .graphql.object_types import MessageObjectType

        return MessageObjectType

    def save(self, *args, **kwargs):
        created = self._state.adding
        super().save(*args, **kwargs)

        if created:
            from baseapp_chats.graphql.subscriptions import (
                ChatRoomOnMessagesCountUpdate,
            )

            for participant in self.room.participants.all():
                if participant.profile_id != self.profile_id:
                    ChatRoomOnMessagesCountUpdate.send_updated_chat_count(
                        profile=participant.profile, profile_id=participant.profile.relay_id
                    )


class AbstractChatRoomParticipant(TimeStampedModel, RelayModel):
    class ChatRoomParticipantRoles(models.IntegerChoices):
        MEMBER = 1, _("member")
        ADMIN = 2, _("admin")

        @property
        def description(self):
            return self.label

    profile = models.ForeignKey(
        swapper.get_model_name("baseapp_profiles", "Profile"),
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    room = models.ForeignKey(
        swapper.get_model_name("baseapp_chats", "ChatRoom"),
        related_name="participants",
        on_delete=models.CASCADE,
    )
    role = models.IntegerField(
        choices=ChatRoomParticipantRoles.choices, default=ChatRoomParticipantRoles.MEMBER
    )
    accepted_at = models.DateTimeField(null=True, blank=True)
    has_archived_room = models.BooleanField(default=False)

    class Meta:
        abstract = True
        ordering = ["-role", "profile__name"]

    @classmethod
    def get_graphql_object_type(cls):
        from .graphql.object_types import ChatRoomParticipantObjectType

        return ChatRoomParticipantObjectType


class AbstractUnreadMessageCount(RelayModel):
    room = models.ForeignKey(
        swapper.get_model_name("baseapp_chats", "ChatRoom"),
        related_name="unread_messages",
        on_delete=models.CASCADE,
    )
    profile = models.ForeignKey(
        swapper.get_model_name("baseapp_profiles", "Profile"), on_delete=models.CASCADE
    )
    marked_unread = models.BooleanField(default=False)
    count = models.IntegerField(default=0)

    class Meta:
        abstract = True
        ordering = ["id"]
        unique_together = ["room_id", "profile_id"]


class AbstractMessageStatus(RelayModel):
    message = models.ForeignKey(
        swapper.get_model_name("baseapp_chats", "Message"),
        on_delete=models.CASCADE,
        related_name="statuses",
    )
    profile = models.ForeignKey(
        swapper.get_model_name("baseapp_profiles", "Profile"), on_delete=models.CASCADE
    )
    is_read = models.BooleanField(default=False)
    read_at = models.DateTimeField(null=True, blank=True)

    class Meta:
        abstract = True
        unique_together = ["message_id", "profile_id"]



import swapper
from django.contrib import admin

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")
Message = swapper.load_model("baseapp_chats", "Message")
MessageStatus = swapper.load_model("baseapp_chats", "MessageStatus")


class ChatRoomParticipantInline(admin.TabularInline):
    model = ChatRoomParticipant
    extra = 0


@admin.register(ChatRoomParticipant)
class ChatRoomParticipantAdmin(admin.ModelAdmin):
    list_display = ("profile", "room", "role")
    list_filter = ["created", "role"]
    search_fields = ["profile", "room"]


class MessageStatusInline(admin.TabularInline):
    model = MessageStatus
    extra = 0


@admin.register(Message)
class MessageAdmin(admin.ModelAdmin):
    list_display = ("profile", "verb", "action_object", "room", "timesince")
    list_filter = ["verb"]
    search_fields = ["content", "profile", "action_object", "room"]
    inlines = [MessageStatusInline]


@admin.register(ChatRoom)
class ChatRoomAdmin(admin.ModelAdmin):
    list_display = ("room",)
    search_fields = ["room"]
    inlines = [ChatRoomParticipantInline]

    def room(self, obj):
        return obj.__str__()



import swapper

from .base import (
    AbstractBaseChatRoom,
    AbstractBaseMessage,
    AbstractChatRoomParticipant,
    AbstractMessageStatus,
    AbstractUnreadMessageCount,
)
from .triggers import (
    create_message_status_trigger,
    decrement_unread_count_trigger,
    increment_unread_count_trigger,
    set_last_message_on_insert_trigger,
    update_last_message_on_delete_trigger,
)


class ChatRoom(AbstractBaseChatRoom):
    class Meta(AbstractBaseChatRoom.Meta):
        swappable = swapper.swappable_setting("baseapp_chats", "ChatRoom")


class ChatRoomParticipant(AbstractChatRoomParticipant):
    class Meta(AbstractChatRoomParticipant.Meta):
        swappable = swapper.swappable_setting("baseapp_chats", "ChatRoomParticipant")


class UnreadMessageCount(AbstractUnreadMessageCount):
    class Meta(AbstractUnreadMessageCount.Meta):
        swappable = swapper.swappable_setting("baseapp_chats", "UnreadMessageCount")


class Message(AbstractBaseMessage):
    class Meta(AbstractBaseMessage.Meta):
        swappable = swapper.swappable_setting("baseapp_chats", "Message")
        triggers = (
            []
            if swapper.is_swapped("baseapp_chats", "Message")
            else [
                set_last_message_on_insert_trigger(ChatRoom),
                create_message_status_trigger(ChatRoomParticipant, AbstractBaseMessage.MessageType),
                update_last_message_on_delete_trigger(ChatRoom),
            ]
        )


class MessageStatus(AbstractMessageStatus):
    class Meta(AbstractMessageStatus.Meta):
        swappable = swapper.swappable_setting("baseapp_chats", "MessageStatus")
        triggers = (
            []
            if swapper.is_swapped("baseapp_chats", "MessageStatus")
            else [
                increment_unread_count_trigger(UnreadMessageCount, Message),
                decrement_unread_count_trigger(UnreadMessageCount, Message),
            ]
        )



import pytest
import swapper

from .factories import ChatRoomFactory, ChatRoomParticipantFactory, MessageFactory

pytestmark = pytest.mark.django_db

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
Message = swapper.load_model("baseapp_chats", "Message")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")
UnreadMessageCount = swapper.load_model("baseapp_chats", "UnreadMessageCount")
MessageStatus = swapper.load_model("baseapp_chats", "MessageStatus")


def test_last_message_is_set():
    room = ChatRoomFactory()
    message = MessageFactory(room=room)

    room.refresh_from_db()
    assert room.last_message == message
    assert room.last_message_time == message.created


def test_last_message_deleted_updates_to_previous():
    room = ChatRoomFactory()
    old_message = MessageFactory(room=room)
    new_message = MessageFactory(room=room)

    room.refresh_from_db()
    assert room.last_message_id == new_message.pk
    assert room.last_message_time == new_message.created

    new_message.delete()
    room.refresh_from_db()
    assert room.last_message_id == old_message.pk
    assert room.last_message_time == old_message.created


def test_message_status_are_created():
    room = ChatRoomFactory()
    participants_count = 2
    ChatRoomParticipantFactory.create_batch(participants_count, room=room)

    message = MessageFactory(room=room)

    assert MessageStatus.objects.filter(message=message).count() == participants_count


def test_does_not_create_message_status_for_system_messages():
    room = ChatRoomFactory()
    participants_count = 2
    ChatRoomParticipantFactory.create_batch(participants_count, room=room)

    message = MessageFactory(room=room, message_type=Message.MessageType.SYSTEM_GENERATED)

    assert MessageStatus.objects.filter(message=message).count() == 0


def test_unread_message_count_is_incremented():
    room = ChatRoomFactory()
    participants_count = 2
    participants = ChatRoomParticipantFactory.create_batch(participants_count, room=room)

    MessageFactory(room=room)

    for participant in participants:
        assert UnreadMessageCount.objects.get(profile=participant.profile).count == 1


def test_unread_message_count_is_not_incremented_for_system_messages():
    room = ChatRoomFactory()
    participants_count = 2
    participants = ChatRoomParticipantFactory.create_batch(participants_count, room=room)

    MessageFactory(room=room, message_type=Message.MessageType.SYSTEM_GENERATED)

    for participant in participants:
        assert UnreadMessageCount.objects.filter(profile=participant.profile).count() == 0


def test_unread_message_count_is_decremented():
    room = ChatRoomFactory()
    participants_count = 2
    participants = ChatRoomParticipantFactory.create_batch(participants_count, room=room)

    message = MessageFactory(room=room)

    message.statuses.all().update(is_read=True)

    for participant in participants:
        assert UnreadMessageCount.objects.get(profile=participant.profile).count == 0



import channels_graphql_ws
import graphene
import swapper
from baseapp_core.graphql import get_obj_from_relay_id
from channels.db import database_sync_to_async

Profile = swapper.load_model("baseapp_profiles", "Profile")
ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")
Message = swapper.load_model("baseapp_chats", "Message")
MessageObjectType = Message.get_graphql_object_type()
ProfileObjectType = Profile.get_graphql_object_type()
ChatRoomObjectType = ChatRoom.get_graphql_object_type()
ChatRoomParticipantObjectType = ChatRoomParticipant.get_graphql_object_type()


class ChatRoomOnRoomUpdate(channels_graphql_ws.Subscription):
    room = graphene.Field(ChatRoomObjectType._meta.connection.Edge)
    removed_participants = graphene.List(ChatRoomParticipantObjectType)

    class Arguments:
        profile_id = graphene.ID(required=True)

    @staticmethod
    async def subscribe(root, info, profile_id):
        user = info.context.channels_scope["user"]
        profile = await database_sync_to_async(get_obj_from_relay_id)(info, profile_id)
        if not user.is_authenticated or not database_sync_to_async(user.has_perm)(
            "baseapp_profiles.use_profile", profile
        ):
            return []
        return [str(profile.pk)]

    @staticmethod
    def publish(payload, info, profile_id):
        return ChatRoomOnRoomUpdate(
            room=ChatRoomObjectType._meta.connection.Edge(node=payload["room"]),
            removed_participants=payload["removed_participants"],
        )

    @classmethod
    def new_message(cls, message):
        cls.room_updated(message.room)

    @classmethod
    def room_updated(cls, room, removed_participants=[]):
        participant_ids = list(room.participants.values_list("profile_id", flat=True))
        removed_ids = [participant.profile.id for participant in removed_participants]
        for id in participant_ids + removed_ids:
            cls.broadcast(
                group=str(id),
                payload={"room": room, "removed_participants": removed_participants},
            )


class ChatRoomOnMessagesCountUpdate(channels_graphql_ws.Subscription):
    profile = graphene.Field(ProfileObjectType)

    class Arguments:
        profile_id = graphene.ID(required=True)

    @staticmethod
    def subscribe(root, info, profile_id):
        user = info.context.channels_scope["user"]
        profile = database_sync_to_async(get_obj_from_relay_id)(info, profile_id)

        # TO DO: change to a better permission check, maybe baseapp_chats.view_chatroom
        if not user.is_authenticated or not database_sync_to_async(user.has_perm)(
            "baseapp_profiles.use_profile", profile
        ):
            return []
        return [profile_id]

    @staticmethod
    def publish(payload, info, profile_id):
        profile = payload["profile"]

        return ChatRoomOnMessagesCountUpdate(profile=profile)

    @classmethod
    def send_updated_chat_count(cls, profile, profile_id):
        cls.broadcast(
            group=profile_id,
            payload={"profile": profile},
        )


class ChatRoomOnNewMessage(channels_graphql_ws.Subscription):
    message = graphene.Field(lambda: MessageObjectType._meta.connection.Edge)

    class Arguments:
        room_id = graphene.ID(required=True)

    @staticmethod
    def subscribe(root, info, room_id):
        room = database_sync_to_async(get_obj_from_relay_id)(info, room_id)

        user = info.context.channels_scope["user"]
        if not user.is_authenticated or not database_sync_to_async(user.has_perm)(
            "baseapp_chats.view_chatroom", room
        ):
            return []
        return [room_id]

    @staticmethod
    def publish(payload, info, room_id):
        message = payload["message"]
        user = info.context.channels_scope["user"]

        if not user.is_authenticated:
            return None

        return ChatRoomOnNewMessage(message=MessageObjectType._meta.connection.Edge(node=message))

    @classmethod
    def new_message(cls, message, room_id):
        cls.broadcast(
            group=room_id,
            payload={"message": message},
        )
        ChatRoomOnRoomUpdate.new_message(message=message)


class ChatsSubscriptions:
    chat_room_on_new_message = ChatRoomOnNewMessage.Field()
    chat_room_on_room_update = ChatRoomOnRoomUpdate.Field()
    chat_room_on_messages_count_update = ChatRoomOnMessagesCountUpdate.Field()



import graphene
import swapper
from baseapp_core.graphql import (
    DjangoObjectType,
    ThumbnailField,
    get_object_type_for_model,
    get_pk_from_relay_id,
)
from django.db.models import Case, When
from graphene import relay
from graphene_django import DjangoConnectionField
from graphene_django.filter import DjangoFilterConnectionField

from .filters import ChatRoomFilter

Profile = swapper.load_model("baseapp_profiles", "Profile")
ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")
ChatRoomParticipantRoleTypesEnum = graphene.Enum.from_enum(
    ChatRoomParticipant.ChatRoomParticipantRoles
)
UnreadMessageCount = swapper.load_model("baseapp_chats", "UnreadMessageCount")
Message = swapper.load_model("baseapp_chats", "Message")


class BaseChatRoomParticipantObjectType:
    role = graphene.Field(ChatRoomParticipantRoleTypesEnum)

    class Meta:
        interfaces = (relay.Node,)
        model = ChatRoomParticipant
        fields = ("id", "has_archived_room", "profile", "role")
        filter_fields = ("profile__target_content_type",)


class ChatRoomParticipantObjectType(BaseChatRoomParticipantObjectType, DjangoObjectType):
    class Meta(BaseChatRoomParticipantObjectType.Meta):
        pass


VerbsEnum = graphene.Enum.from_enum(Message.Verbs)
MessageTypeEnum = graphene.Enum.from_enum(Message.MessageType)


class BaseMessageObjectType:
    action_object = graphene.Field(relay.Node)
    verb = graphene.Field(VerbsEnum)
    message_type = graphene.Field(MessageTypeEnum)
    content = graphene.String(required=False, profile_id=graphene.ID(required=False))
    is_read = graphene.Boolean(profile_id=graphene.ID(required=False))

    class Meta:
        interfaces = (relay.Node,)
        model = Message
        fields = (
            "id",
            "verb",
            "content",
            "message_type",
            "user",
            "profile",
            "created",
            "room",
            "action_object",
            "extra_data",
            "in_reply_to",
            "is_read",
        )
        filter_fields = ("verb",)

    @classmethod
    def get_node(cls, info, id):
        try:
            message = cls._meta.model.objects.get(id=id)
            if not info.context.user.has_perm("baseapp_chats.view_message", message):
                return None
            return message
        except cls._meta.model.DoesNotExist:
            return None

    @staticmethod
    def get_profile_pk(info, profile_id=None):
        if profile_id:
            profile_pk = get_pk_from_relay_id(profile_id)
            if not Profile.objects.get_if_member(pk=profile_pk, user=info.context.user):
                return None
            else:
                return profile_pk
        elif hasattr(info.context.user, "current_profile") and hasattr(
            info.context.user.current_profile, "pk"
        ):
            return info.context.user.current_profile.pk
        elif hasattr(info.context.user, "profile") and hasattr(info.context.user.profile, "pk"):
            return info.context.user.profile.pk
        else:
            return None

    @staticmethod
    def get_replaced_profile_name(profile, profile_pk, replacement_text):
        if not profile:
            return None
        elif profile.id == profile_pk:
            return replacement_text
        else:
            return profile.name

    @staticmethod
    def resolve_content(root, info, profile_id=None, **kwargs):
        profile_pk = BaseMessageObjectType.get_profile_pk(info, profile_id)
        if not profile_pk:
            return None
        if root.message_type == Message.MessageType.USER_MESSAGE:
            return root.content

        linked_capital_name = BaseMessageObjectType.get_replaced_profile_name(
            root.content_linked_profile_actor, profile_pk, "You"
        )
        linked_small_name = BaseMessageObjectType.get_replaced_profile_name(
            root.content_linked_profile_target, profile_pk, "you"
        )
        return root.content.format(
            content_linked_profile_actor=linked_capital_name,
            content_linked_profile_target=linked_small_name,
        )

    @staticmethod
    def resolve_is_read(root, info, profile_id=None, **kwargs):
        profile_pk = BaseMessageObjectType.get_profile_pk(info, profile_id)
        if not profile_pk:
            return None

        message_status = root.statuses.filter(profile_id=profile_pk).first()
        return message_status and message_status.is_read


class MessageObjectType(BaseMessageObjectType, DjangoObjectType):
    class Meta(BaseMessageObjectType.Meta):
        pass


class BaseChatRoomObjectType:
    all_messages = DjangoFilterConnectionField(get_object_type_for_model(Message))
    participants = DjangoConnectionField(get_object_type_for_model(ChatRoomParticipant))
    unread_messages = graphene.Field(
        get_object_type_for_model(UnreadMessageCount), profile_id=graphene.ID(required=False)
    )
    image = ThumbnailField(required=False)
    is_archived = graphene.Boolean(profile_id=graphene.ID(required=False))

    @classmethod
    def get_node(cls, info, id):
        try:
            room = cls._meta.model.objects.get(id=id)
            if not info.context.user.has_perm("baseapp_chats.view_chatroom", room):
                return None
            return room

        except cls._meta.model.DoesNotExist:
            return None

    def resolve_all_messages(self, info, **kwargs):
        if self.is_group:
            profile = (
                info.context.user.current_profile
                if hasattr(info.context.user, "current_profile")
                else (
                    info.context.user.profile.pk if hasattr(info.context.user, "profile") else None
                )
            )
            participant = self.participants.filter(profile=profile).first()
            if participant:
                return self.messages.filter(created__gte=participant.accepted_at).order_by(
                    "-created"
                )
        return self.messages.all().order_by("-created")

    def resolve_participants(self, info, **kwargs):
        if self.is_group:
            profile = (
                info.context.user.current_profile
                if hasattr(info.context.user, "current_profile")
                else (info.context.user.profile if hasattr(info.context.user, "profile") else None)
            )
            participant = self.participants.filter(profile=profile).first()
            if participant:
                return self.participants.all().order_by(
                    Case(When(id=participant.id, then=0), default=1),
                    "-role",
                    "profile__name",
                )
        return self.participants.all()

    def resolve_is_archived(self, info, profile_id=None, **kwargs):
        if profile_id:
            profile_pk = get_pk_from_relay_id(profile_id)
            profile = Profile.objects.get_if_member(pk=profile_pk, user=info.context.user)
            if not profile:
                return None
        else:
            profile_pk = (
                info.context.user.current_profile.pk
                if hasattr(info.context.user, "current_profile")
                and hasattr(info.context.user.current_profile, "pk")
                else (
                    info.context.user.profile.pk
                    if hasattr(info.context.user, "profile")
                    and hasattr(info.context.user.profile, "pk")
                    else None
                )
            )
        return self.participants.filter(profile_id=profile_pk, has_archived_room=True).exists()

    def resolve_unread_messages(self, info, profile_id=None, **kwargs):
        if profile_id:
            profile_pk = get_pk_from_relay_id(profile_id)
            profile = Profile.objects.get_if_member(pk=profile_pk, user=info.context.user)
            if not profile:
                return None
        else:
            profile_pk = (
                info.context.user.current_profile
                if hasattr(info.context.user, "current_profile")
                and hasattr(info.context.user.current_profile, "pk")
                else (
                    info.context.user.profile.pk
                    if hasattr(info.context.user, "profile")
                    and hasattr(info.context.user.profile, "pk")
                    else None
                )
            )

        unread_messages = UnreadMessageCount.objects.filter(
            room=self,
            profile_id=profile_pk,
        ).first()

        return unread_messages

    class Meta:
        interfaces = (relay.Node,)
        model = ChatRoom
        fields = (
            "id",
            "last_message_time",
            "last_message",
            "participants",
            "participants_count",
            "title",
            "image",
            "is_group",
        )
        filterset_class = ChatRoomFilter


class ChatRoomObjectType(BaseChatRoomObjectType, DjangoObjectType):
    class Meta(BaseChatRoomObjectType.Meta):
        pass


class BaseUnreadMessageObjectType:
    class Meta:
        interfaces = (relay.Node,)
        model = UnreadMessageCount
        fields = ("count", "marked_unread")


class UnreadMessageObjectType(BaseUnreadMessageObjectType, DjangoObjectType):
    class Meta(BaseUnreadMessageObjectType.Meta):
        pass



# Generated by Django 5.0.8 on 2025-01-21 11:57

import django.db.models.deletion
import pgtrigger.compiler
import pgtrigger.migrations
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("baseapp_chats", "0008_alter_chatroomparticipant_options"),
        migrations.swappable_dependency(settings.BASEAPP_PROFILES_PROFILE_MODEL),
    ]

    operations = [
        pgtrigger.migrations.RemoveTrigger(
            model_name="message",
            name="create_message_status",
        ),
        migrations.AddField(
            model_name="message",
            name="content_linked_profile_actor",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="linked_as_content_actor_set",
                to=settings.BASEAPP_PROFILES_PROFILE_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="message",
            name="content_linked_profile_target",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="linked_as_content_target_set",
                to=settings.BASEAPP_PROFILES_PROFILE_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="message",
            name="message_type",
            field=models.IntegerField(
                choices=[(1, "User message"), (2, "System message")], default=1
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="message",
            trigger=pgtrigger.compiler.Trigger(
                name="create_message_status",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (NEW."message_type" = 1)',
                    func="\n            INSERT INTO baseapp_chats_messagestatus (message_id, profile_id, is_read)\n            SELECT NEW.id, crp.profile_id, CASE WHEN NEW.profile_id = crp.profile_id THEN TRUE ELSE FALSE END\n            FROM baseapp_chats_chatroomparticipant as crp\n            WHERE crp.room_id = NEW.room_id;\n            RETURN NULL;\n        ",
                    hash="beb46ab0586aff0156230e06f0f378f3af30b54b",
                    operation="INSERT",
                    pgid="pgtrigger_create_message_status_c941b",
                    table="baseapp_chats_message",
                    when="AFTER",
                ),
            ),
        ),
    ]



# Generated by Django 5.0.9 on 2024-11-13 23:31

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("baseapp_chats", "0002_message_set_last_message_and_more"),
        migrations.swappable_dependency(settings.BASEAPP_CHATS_CHATROOM_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name="unreadmessagecount",
            name="room",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="unread_messages",
                to=settings.BASEAPP_CHATS_CHATROOM_MODEL,
            ),
        ),
    ]



# Generated by Django 5.0.9 on 2024-12-21 16:29

import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("baseapp_chats", "0005_remove_messagestatus_increment_unread_count_and_more"),
    ]

    operations = [
        pgtrigger.migrations.RemoveTrigger(
            model_name="messagestatus",
            name="increment_unread_count",
        ),
        migrations.AddField(
            model_name="chatroom",
            name="is_group",
            field=models.BooleanField(default=False),
        ),
    ]



import graphene
from baseapp_profiles.graphql.queries import ProfilesQueries

from baseapp_chats.graphql.mutations import ChatsMutations
from baseapp_chats.graphql.queries import ChatsQueries
from baseapp_chats.graphql.subscriptions import ChatsSubscriptions
from testproject.testapp.graphql.queries import UsersQueries


class Query(graphene.ObjectType, UsersQueries, ProfilesQueries, ChatsQueries):
    pass


class Mutation(
    graphene.ObjectType,
    ChatsMutations,
):
    pass


class Subscription(graphene.ObjectType, ChatsSubscriptions):
    pass


schema = graphene.Schema(query=Query, mutation=Mutation, subscription=Subscription)



import pgtrigger
from django.db import models
from pgtrigger import utils


class Func(pgtrigger.Func):
    def render(self, model: models.Model) -> str:
        fields = utils.AttrDict({field.name: field for field in model._meta.fields})
        columns = utils.AttrDict({field.name: field.column for field in model._meta.fields})
        return self.func.format(model=model, meta=model._meta, fields=fields, columns=columns)


def increment_unread_count_trigger(UnreadMessageCount, Message):
    return pgtrigger.Trigger(
        name="increment_unread_count",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Insert,
        condition=pgtrigger.Q(new__is_read=False),
        func=f"""
            INSERT INTO {UnreadMessageCount._meta.db_table} (room_id, profile_id, marked_unread, count)
            VALUES ((SELECT room_id FROM {Message._meta.db_table} WHERE id = NEW.message_id), NEW.profile_id, False, 1)
            ON CONFLICT (room_id, profile_id)
            DO UPDATE SET count = {UnreadMessageCount._meta.db_table}.count + 1;
            RETURN NULL;
        """,
    )


def decrement_unread_count_trigger(UnreadMessageCount, Message):
    return pgtrigger.Trigger(
        name="decrement_unread_count",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(new__is_read=True, old__is_read=False),
        func=f"""
            UPDATE {UnreadMessageCount._meta.db_table}
            SET count = GREATEST(0, count - 1)
            WHERE
                room_id = (SELECT room_id FROM {Message._meta.db_table} WHERE id = NEW.message_id) AND
                profile_id = NEW.profile_id;
            RETURN NULL;
        """,
    )


# Create MessageStatus row in the database for each CharRoomParticipant
def create_message_status_trigger(ChatRoomParticipant, MessageType):
    return pgtrigger.Trigger(
        name="create_message_status",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Insert,
        condition=pgtrigger.Q(new__message_type=MessageType.USER_MESSAGE),
        func=Func(
            f"""
            INSERT INTO {{model.statuses.field.model._meta.db_table}} (message_id, profile_id, is_read)
            SELECT NEW.id, crp.profile_id, CASE WHEN NEW.profile_id = crp.profile_id THEN TRUE ELSE FALSE END
            FROM {ChatRoomParticipant._meta.db_table} as crp
            WHERE crp.room_id = NEW.room_id;
            RETURN NULL;
        """
        ),
    )


# Set ChatRoom last_message and last_message_time fields when a new Message is created
def set_last_message_on_insert_trigger(ChatRoom):
    return pgtrigger.Trigger(
        name="set_last_message",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Insert,
        func=f"""
            UPDATE {ChatRoom._meta.db_table}
            SET last_message_id = NEW.id, last_message_time = NEW.created
            WHERE id = NEW.room_id;
            RETURN NULL;
        """,
    )


# Update ChatRoom last_message and last_message_time fields to previous message when the last message is deleted
def update_last_message_on_delete_trigger(ChatRoom):
    return pgtrigger.Trigger(
        name="update_last_message",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Delete,
        func=Func(
            f"""
            UPDATE {ChatRoom._meta.db_table}
            SET last_message_id = (
                SELECT id
                FROM {{meta.db_table}}
                WHERE room_id = OLD.room_id
                ORDER BY created DESC
                LIMIT 1
            ),
            last_message_time = (
                SELECT created
                FROM {{meta.db_table}}
                WHERE room_id = OLD.room_id
                ORDER BY created DESC
                LIMIT 1
            )
            WHERE id = OLD.room_id;
            RETURN NULL;
        """
        ),
    )



import pytest
import swapper
from baseapp_blocks.tests.factories import BlockFactory
from baseapp_core.graphql.testing.fixtures import graphql_query
from baseapp_core.tests.factories import UserFactory
from baseapp_profiles.tests.factories import ProfileFactory
from django.test.client import MULTIPART_CONTENT

from .factories import ChatRoomFactory, ChatRoomParticipantFactory, MessageFactory
from .test_graphql_queries import PROFILE_ROOMS_GRAPHQL, ROOM_GRAPHQL

pytestmark = pytest.mark.django_db

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
Message = swapper.load_model("baseapp_chats", "Message")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")
UnreadMessageCount = swapper.load_model("baseapp_chats", "UnreadMessageCount")
MessageStatus = swapper.load_model("baseapp_chats", "MessageStatus")


SEND_MESSAGE_GRAPHQL = """
    mutation SendMessageMutation($input: ChatRoomSendMessageInput!) {
        chatRoomSendMessage(input: $input) {
            message {
                node {
                    id
                    content
                    inReplyTo {
                        id
                    }
                }
            }
            errors {
                field
                messages
            }
        }
    }
"""

CREATE_ROOM_GRAPHQL = """
    mutation CreateRoomMutation($input: ChatRoomCreateInput!) {
        chatRoomCreate(input: $input) {
            room {
                node {
                    id
                    title
                    isGroup
                    participantsCount
                    participants {
                        edges {
                            node {
                                id
                            }
                        }
                    }
                    image(width: 100, height: 100) {
                        url
                    }
                    allMessages {
                        edges {
                            node {
                                id
                                content
                                messageType
                            }
                        }
                    }
                }
            }
            errors {
                field
                messages
            }
        }
    }
"""

UPDATE_ROOM_GRAPHQL = """
    mutation UpdateRoomMutation($input: ChatRoomUpdateInput!) {
        chatRoomUpdate(input: $input) {
            room {
                node {
                    id
                    title
                    isGroup
                    participantsCount
                    participants {
                        edges {
                            node {
                                id
                                profile {
                                    id
                                }
                            }
                        }
                    }
                    image(width: 100, height: 100) {
                        url
                    }
                }
            }
            removedParticipants {
                id
                profile {
                    id
                }
            }
            errors {
                field
                messages
            }
        }
    }
"""

EDIT_MESSAGE_GRAPHQL = """
    mutation EditMessageMutation($input: ChatRoomEditMessageInput!) {
        chatRoomEditMessage(input: $input) {
            message {
                node {
                    id
                    content
                }
            }
            errors {
                field
                messages
            }
        }
    }
"""

READ_MESSAGE_GRAPHQL = """
    mutation ReadMessageMutation($input: ChatRoomReadMessagesInput!) {
        chatRoomReadMessages(input: $input) {
            room {
                id
                unreadMessages {
                    count
                    markedUnread
                }
            }
            errors {
                field
                messages
            }
        }
    }
"""

UNREAD_CHAT_GRAPHQL = """
    mutation UnreadMutation($input: ChatRoomUnreadInput!) {
        chatRoomUnread(input: $input) {
            room {
                id
                unreadMessages {
                    count
                    markedUnread
                }
            }
            errors {
                field
                messages
            }
        }
    }
"""

ARCHIVE_CHAT_ROOM_GRAPHQL = """
    mutation ChatRoomArchiveMutation($input: ChatRoomArchiveInput!) {
        chatRoomArchive(input: $input) {
            room {
                id
            }
            errors {
                field
                messages
            }
        }
    }
"""


def test_user_can_read_all_messages(graphql_user_client, django_user_client):
    room = ChatRoomFactory(created_by=django_user_client.user)

    my_profile = django_user_client.user.profile

    friend = ProfileFactory()
    ChatRoomParticipantFactory(room=room, profile=my_profile)
    friend_participant = ChatRoomParticipantFactory(room=room, profile=friend)

    MessageFactory.create_batch(2, room=room, profile=my_profile, user=my_profile.owner)
    MessageFactory.create_batch(
        2, room=room, profile=friend_participant.profile, user=friend_participant.profile.owner
    )

    assert MessageStatus.objects.filter(profile=my_profile, is_read=False).count() == 2
    assert (
        MessageStatus.objects.filter(profile=friend_participant.profile, is_read=False).count() == 2
    )

    # makes sure the cached field is increased when a participant send a message
    assert UnreadMessageCount.objects.filter(profile=my_profile, room=room).first().count == 2
    assert (
        UnreadMessageCount.objects.filter(profile=friend_participant.profile, room=room)
        .first()
        .count
        == 2
    )

    response = graphql_user_client(
        READ_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": my_profile.relay_id,
            },
        },
    )
    content = response.json()

    assert content["data"]["chatRoomReadMessages"]["room"]["id"] == room.relay_id
    assert MessageStatus.objects.filter(profile=my_profile, is_read=False).count() == 0
    assert MessageStatus.objects.filter(profile=my_profile, is_read=True).count() == 4
    assert (
        MessageStatus.objects.filter(profile=friend_participant.profile, is_read=False).count() == 2
    )

    assert (
        UnreadMessageCount.objects.filter(profile=friend_participant.profile, room=room)
        .first()
        .count
        == 2
    )
    # makes sure the cached field is decreased when a participant read a message
    assert content["data"]["chatRoomReadMessages"]["room"]["unreadMessages"]["count"] == 0
    assert UnreadMessageCount.objects.filter(profile=my_profile, room=room).first().count == 0


def test_user_can_read_one_message(graphql_user_client, django_user_client):
    room = ChatRoomFactory(created_by=django_user_client.user)

    my_profile = django_user_client.user.profile

    friend = ProfileFactory()
    ChatRoomParticipantFactory(room=room, profile=my_profile)
    friend_participant = ChatRoomParticipantFactory(room=room, profile=friend)

    MessageFactory.create_batch(2, room=room, profile=my_profile, user=my_profile.owner)
    friend_messages = MessageFactory.create_batch(
        2, room=room, profile=friend_participant.profile, user=friend_participant.profile.owner
    )
    message = friend_messages[0]

    assert MessageStatus.objects.filter(profile=my_profile, is_read=False).count() == 2
    assert (
        MessageStatus.objects.filter(profile=friend_participant.profile, is_read=False).count() == 2
    )

    # makes sure the cached field is increased when a participant send a message
    assert UnreadMessageCount.objects.filter(profile=my_profile, room=room).first().count == 2
    assert (
        UnreadMessageCount.objects.filter(profile=friend_participant.profile, room=room)
        .first()
        .count
        == 2
    )

    response = graphql_user_client(
        READ_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": my_profile.relay_id,
                "messageIds": [message.relay_id],
            },
        },
    )
    content = response.json()

    assert content["data"]["chatRoomReadMessages"]["room"]["id"] == room.relay_id
    assert MessageStatus.objects.filter(profile=my_profile, is_read=False).count() == 1
    assert MessageStatus.objects.filter(profile=my_profile, is_read=True).count() == 3
    assert (
        MessageStatus.objects.filter(profile=friend_participant.profile, is_read=False).count() == 2
    )

    assert (
        UnreadMessageCount.objects.filter(profile=friend_participant.profile, room=room)
        .first()
        .count
        == 2
    )
    # makes sure the cached field is decreased when a participant read a message
    assert content["data"]["chatRoomReadMessages"]["room"]["unreadMessages"]["count"] == 1
    assert UnreadMessageCount.objects.filter(profile=my_profile, room=room).first().count == 1


def test_user_can_unread_chat(graphql_user_client, django_user_client):
    room = ChatRoomFactory(created_by=django_user_client.user)

    my_profile = django_user_client.user.profile

    ChatRoomParticipantFactory(room=room, profile=my_profile)
    MessageFactory.create_batch(2, room=room, profile=my_profile, user=my_profile.owner)

    # Unread chat and check it is marked unread
    response = graphql_user_client(
        UNREAD_CHAT_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": my_profile.relay_id,
            },
        },
    )
    content = response.json()
    assert content["data"]["chatRoomUnread"]["room"]["id"] == room.relay_id
    assert content["data"]["chatRoomUnread"]["room"]["unreadMessages"]["markedUnread"] is True
    assert (
        UnreadMessageCount.objects.filter(profile=my_profile, room=room).first().marked_unread
        is True
    )

    # Read chat and check the unread mark is removed
    response = graphql_user_client(
        READ_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": my_profile.relay_id,
            },
        },
    )
    content = response.json()
    assert content["data"]["chatRoomReadMessages"]["room"]["id"] == room.relay_id
    assert (
        content["data"]["chatRoomReadMessages"]["room"]["unreadMessages"]["markedUnread"] is False
    )
    assert (
        UnreadMessageCount.objects.filter(profile=my_profile, room=room).first().marked_unread
        is False
    )


@pytest.mark.celery_app
def test_user_can_send_message(django_user_client, graphql_user_client, celery_config):
    user = django_user_client.user
    room = ChatRoomFactory(created_by=user)
    friend = ProfileFactory()

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=friend, room=room)

    response = graphql_user_client(
        SEND_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": user.profile.relay_id,
                "content": "blablabla",
            }
        },
    )

    content = response.json()

    assert content["data"]["chatRoomSendMessage"]["message"]["node"]["content"] == "blablabla"
    # check if message status is created for both participants
    assert MessageStatus.objects.filter(profile=user.profile, is_read=False).count() == 0
    assert MessageStatus.objects.filter(profile=friend, is_read=False).count() == 1
    room.refresh_from_db()
    assert room.messages.count() == 1


@pytest.mark.celery_app
def test_user_can_send_message_in_reply_to(django_user_client, graphql_user_client, celery_config):
    user = django_user_client.user
    room = ChatRoomFactory(created_by=user)
    friend = ProfileFactory()

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=friend, room=room)

    message = MessageFactory(room=room, profile=friend, user=friend.owner)

    response = graphql_user_client(
        SEND_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": user.profile.relay_id,
                "content": "blablabla",
                "inReplyToId": message.relay_id,
            }
        },
    )

    content = response.json()
    assert (
        content["data"]["chatRoomSendMessage"]["message"]["node"]["inReplyTo"]["id"]
        == message.relay_id
    )


def test_cant_send_message_non_participating_room(django_user_client, graphql_user_client):
    user = UserFactory()
    room = ChatRoomFactory(created_by=user)
    ChatRoomParticipantFactory(profile=user.profile, room=room)

    response = graphql_user_client(
        SEND_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": django_user_client.user.profile.relay_id,
                "content": "blablabla",
            }
        },
    )

    content = response.json()

    assert (
        content["data"]["chatRoomSendMessage"]["errors"][0]["messages"][0]
        == "You don't have permission to send a message in this room"
    )
    room.refresh_from_db()
    assert room.messages.count() == 0


def test_user_can_edit_message(django_user_client, graphql_user_client):
    user = django_user_client.user
    room = ChatRoomFactory(created_by=user)
    friend = ProfileFactory()

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=friend, room=room)

    message = MessageFactory(room=room, profile=user.profile, user=user)

    response = graphql_user_client(
        EDIT_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "id": message.relay_id,
                "content": "edited",
            }
        },
    )

    content = response.json()

    assert content["data"]["chatRoomEditMessage"]["message"]["node"]["content"] == "edited"


def test_user_cant_edit_to_empty_message(django_user_client, graphql_user_client):
    user = django_user_client.user
    room = ChatRoomFactory(created_by=user)
    friend = ProfileFactory()

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=friend, room=room)

    message = MessageFactory(room=room, profile=user.profile, user=user)

    response = graphql_user_client(
        EDIT_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "id": message.relay_id,
                "content": "",
            }
        },
    )

    content = response.json()

    assert (
        content["data"]["chatRoomEditMessage"]["errors"][0]["messages"][0]
        == "You cannot edit an empty message"
    )


def test_user_cant_edit_to_more_than_1000_caracters(django_user_client, graphql_user_client):
    user = django_user_client.user
    room = ChatRoomFactory(created_by=user)
    friend = ProfileFactory()

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=friend, room=room)

    message = MessageFactory(room=room, profile=user.profile, user=user)

    response = graphql_user_client(
        EDIT_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "id": message.relay_id,
                "content": "a" * 1001,
            }
        },
    )

    content = response.json()

    assert (
        content["data"]["chatRoomEditMessage"]["errors"][0]["messages"][0]
        == "Message must be no longer than 1000 characters"
    )


def test_user_cant_edit_a_message_that_does_not_exist(django_user_client, graphql_user_client):
    user = django_user_client.user
    response = graphql_user_client(
        EDIT_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "id": user.relay_id,
                "content": "edited",
            }
        },
    )

    content = response.json()
    print("content", content)
    assert (
        content["data"]["chatRoomEditMessage"]["errors"][0]["messages"][0]
        == "Message does not exist"
    )


def test_current_profile_connot_edit_message_sent_by_another_profile(
    django_user_client, graphql_user_client
):
    user = django_user_client.user
    room = ChatRoomFactory(created_by=user)
    friend = ProfileFactory()

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=friend, room=room)

    message = MessageFactory(room=room, profile=friend, user=friend.owner)

    response = graphql_user_client(
        EDIT_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "id": message.relay_id,
                "content": "edited",
            }
        },
    )

    content = response.json()

    assert (
        content["data"]["chatRoomEditMessage"]["errors"][0]["messages"][0]
        == "You don't have permission to update this message"
    )


def test_user_can_create_room(django_user_client, graphql_user_client):
    band = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "participants": [
                    band.relay_id,
                ],
            }
        },
    )

    content = response.json()

    assert content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    assert len(content["data"]["chatRoomCreate"]["room"]["node"]["participants"]["edges"]) == 2


@pytest.mark.skip(reason="in-app notification is currently turned off")
@pytest.mark.celery_app
def test_user_get_new_message_in_app_notification(
    django_user_client, graphql_user_client, celery_config
):
    room = ChatRoomFactory(created_by=django_user_client.user)
    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    friend = ChatRoomParticipantFactory(room=room)

    graphql_user_client(
        SEND_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": django_user_client.user.profile.relay_id,
                "content": "blablabla",
            }
        },
    )

    notification = friend.profile.owner.notifications.get()
    assert notification.verb == "NEW_CHAT_MESSAGE"


def test_user_cant_send_message_to_blocked_user(django_user_client, graphql_user_client):
    room = ChatRoomFactory(created_by=django_user_client.user)
    user = UserFactory()

    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    ChatRoomParticipantFactory(profile=user.profile, room=room)

    BlockFactory(actor=user.profile, target=django_user_client.user.profile)

    response = graphql_user_client(
        SEND_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": django_user_client.user.profile.relay_id,
                "content": "blablabla",
            }
        },
    )

    content = response.json()
    assert (
        content["data"]["chatRoomSendMessage"]["errors"][0]["messages"][0]
        == "You don't have permission to send a message in this room"
    )
    assert room.messages.count() == 0


def test_blocked_user_cant_send_message_to_user(django_user_client, graphql_user_client):
    room = ChatRoomFactory(created_by=django_user_client.user)
    user = UserFactory()

    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    ChatRoomParticipantFactory(profile=user.profile, room=room)

    BlockFactory(actor=django_user_client.user.profile, target=user.profile)

    response = graphql_user_client(
        SEND_MESSAGE_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": django_user_client.user.profile.relay_id,
                "content": "blablabla",
            }
        },
    )

    content = response.json()
    assert (
        content["data"]["chatRoomSendMessage"]["errors"][0]["messages"][0]
        == "You don't have permission to send a message in this room"
    )
    assert room.messages.count() == 0


def test_user_cant_create_room_with_blocked_user(django_user_client, graphql_user_client):
    user = UserFactory()

    BlockFactory(actor=django_user_client.user.profile, target=user.profile)

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "participants": [
                    user.profile.relay_id,
                ],
            }
        },
    )

    content = response.json()

    assert (
        content["data"]["chatRoomCreate"]["errors"][0]["messages"][0]
        == "You can't create a chatroom with those participants"
    )


def test_blocked_user_cant_create_room_with_user(django_user_client, graphql_user_client):
    friend = ProfileFactory()

    BlockFactory(actor=friend, target=django_user_client.user.profile)

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "participants": [
                    friend.relay_id,
                ],
            }
        },
    )

    content = response.json()

    # since we use get_obj_from_relay_id which uses ObjectType.get_node it will return None for the participant that is blocked
    assert (
        content["data"]["chatRoomCreate"]["errors"][0]["messages"][0]
        == "You can't create a chatroom with those participants"
    )


@pytest.mark.celery_app
def test_user_can_archive_chatroom(django_user_client, graphql_user_client, celery_config):
    user = django_user_client.user
    room = ChatRoomFactory(created_by=user)
    friend = ProfileFactory()

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=friend, room=room)

    # Archive chatroom and query to verify
    response = graphql_user_client(
        ARCHIVE_CHAT_ROOM_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": user.profile.relay_id,
                "archive": True,
            }
        },
    )

    content = response.json()

    assert content["data"]["chatRoomArchive"]["room"]["id"] == room.relay_id

    response = graphql_user_client(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": user.profile.relay_id, "archived": True},
    )

    content = response.json()

    assert len(content["data"]["profile"]["chatRooms"]["edges"]) == 1

    # Unarchive chatroom and query to verify
    graphql_user_client(
        ARCHIVE_CHAT_ROOM_GRAPHQL,
        variables={
            "input": {
                "roomId": room.relay_id,
                "profileId": user.profile.relay_id,
                "archive": False,
            }
        },
    )

    response = graphql_user_client(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": user.profile.relay_id, "archived": True},
    )

    content = response.json()

    assert len(content["data"]["profile"]["chatRooms"]["edges"]) == 0


def test_user_can_create_group(django_user_client, graphql_user_client, image_djangofile):
    participant = ProfileFactory()
    participant_2 = ProfileFactory()
    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    participant.relay_id,
                    participant_2.relay_id,
                ],
            }
        },
        content_type=MULTIPART_CONTENT,
        extra={"image": image_djangofile},
    )

    content = response.json()

    assert content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    assert content["data"]["chatRoomCreate"]["room"]["node"]["title"] == "group"
    assert content["data"]["chatRoomCreate"]["room"]["node"]["isGroup"]
    assert len(content["data"]["chatRoomCreate"]["room"]["node"]["participants"]["edges"]) == 3
    assert content["data"]["chatRoomCreate"]["room"]["node"]["image"]["url"].startswith("http://")


def test_user_cant_create_group_without_title(django_user_client, graphql_user_client):
    friend = ProfileFactory()
    friend_2 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "participants": [
                    friend.relay_id,
                    friend_2.relay_id,
                ],
            }
        },
    )

    content = response.json()

    assert (
        content["data"]["chatRoomCreate"]["errors"][0]["messages"][0]
        == "Title is required for group chats"
    )


def test_create_room_handles_corrupted_images(
    django_user_client, graphql_user_client, corrupted_image
):
    friend = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "participants": [
                    friend.relay_id,
                ],
            }
        },
        content_type=MULTIPART_CONTENT,
        extra={"image": corrupted_image},
    )

    content = response.json()
    assert "corrupted" in content["data"]["chatRoomCreate"]["errors"][0]["messages"][0]


def test_create_room_creates_system_message(django_user_client, graphql_user_client, django_client):
    friend = UserFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "participants": [
                    friend.profile.relay_id,
                ],
                "isGroup": True,
                "title": "A test group",
            }
        },
    )

    content = response.json()
    assert len(content["data"]["chatRoomCreate"]["room"]["node"]["allMessages"]["edges"]) == 1
    assert (
        content["data"]["chatRoomCreate"]["room"]["node"]["allMessages"]["edges"][0]["node"][
            "messageType"
        ]
        == "SYSTEM_GENERATED"
    )
    assert (
        content["data"]["chatRoomCreate"]["room"]["node"]["allMessages"]["edges"][0]["node"][
            "content"
        ]
        == 'You created group "A test group"'
    )
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]

    django_client.force_login(friend)
    response = graphql_query(
        ROOM_GRAPHQL,
        variables={"profileId": friend.profile.relay_id, "roomId": room_id},
        client=django_client,
    )

    content = response.json()
    assert len(content["data"]["chatRoom"]["allMessages"]["edges"]) == 1
    assert (
        content["data"]["chatRoom"]["allMessages"]["edges"][0]["node"]["messageType"]
        == "SYSTEM_GENERATED"
    )
    assert (
        content["data"]["chatRoom"]["allMessages"]["edges"][0]["node"]["content"]
        == django_user_client.user.profile.name + ' created group "A test group"'
    )


def test_member_user_cannot_update_group(django_user_client, graphql_user_client, django_client):
    friend = ProfileFactory()
    friend_2 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend.relay_id,
                    friend_2.relay_id,
                ],
            }
        },
    )
    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    django_client.force_login(friend.owner)

    response = graphql_query(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": friend.relay_id,
                "roomId": room_id,
                "title": "new group",
            }
        },
        client=django_client,
    )
    content = response.json()
    assert (
        content["data"]["chatRoomUpdate"]["errors"][0]["messages"][0]
        == "You don't have permission to update this room"
    )


def test_admin_user_can_update_group_title(django_user_client, graphql_user_client):
    friend = ProfileFactory()
    friend_2 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend.relay_id,
                    friend_2.relay_id,
                ],
            }
        },
    )
    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    assert content["data"]["chatRoomCreate"]["room"]["node"]["title"] == "group"

    response = graphql_user_client(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "roomId": room_id,
                "title": "new group",
            }
        },
    )
    content = response.json()
    assert content["data"]["chatRoomUpdate"]["room"]["node"]["title"] == "new group"


def test_user_cannot_update_room_title(django_user_client, graphql_user_client):
    # user can't update the title of a room that is not a group
    friend = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": False,
                "title": "group",
                "participants": [
                    friend.relay_id,
                ],
            }
        },
    )

    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]

    response = graphql_user_client(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "roomId": room_id,
                "title": "test",
            }
        },
    )
    content = response.json()
    assert (
        content["data"]["chatRoomUpdate"]["errors"][0]["messages"][0]
        == "This room cannot be updated"
    )


def test_admin_user_can_update_group_image(
    django_user_client, graphql_user_client, image_djangofile
):
    friend = ProfileFactory()
    friend_2 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend.relay_id,
                    friend_2.relay_id,
                ],
            }
        },
    )

    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    assert content["data"]["chatRoomCreate"]["room"]["node"]["image"] is None

    response = graphql_user_client(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "roomId": room_id,
            }
        },
        content_type=MULTIPART_CONTENT,
        extra={"image": image_djangofile},
    )
    content = response.json()
    assert content["data"]["chatRoomUpdate"]["room"]["node"]["image"]["url"].startswith("http://")


def test_admin_user_can_delete_group_image(
    django_user_client, graphql_user_client, image_djangofile
):
    # user can delete the image of a group if he is the creator/admin
    friend = ProfileFactory()
    friend_2 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend.relay_id,
                    friend_2.relay_id,
                ],
            }
        },
        content_type=MULTIPART_CONTENT,
        extra={"image": image_djangofile},
    )

    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    assert content["data"]["chatRoomCreate"]["room"]["node"]["image"] is not None

    response = graphql_user_client(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "roomId": room_id,
                "deleteImage": True,
            }
        },
    )
    content = response.json()
    assert content["data"]["chatRoomUpdate"]["room"]["node"]["image"] is None


def test_update_room_handles_corrupted_images(
    django_user_client, graphql_user_client, corrupted_image
):
    friend = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend.relay_id,
                ],
            }
        },
    )

    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]

    response = graphql_user_client(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "roomId": room_id,
            }
        },
        content_type=MULTIPART_CONTENT,
        extra={"image": corrupted_image},
    )

    content = response.json()
    assert "corrupted" in content["data"]["chatRoomUpdate"]["errors"][0]["messages"][0]


def test_admin_user_can_remove_participants(django_user_client, graphql_user_client):
    friend_1 = ProfileFactory()
    friend_2 = ProfileFactory()
    friend_3 = ProfileFactory()
    friend_4 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend_1.relay_id,
                    friend_2.relay_id,
                    friend_3.relay_id,
                    friend_4.relay_id,
                ],
            }
        },
    )
    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    assert content["data"]["chatRoomCreate"]["room"]["node"]["participantsCount"] == 5

    response = graphql_user_client(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "roomId": room_id,
                "removeParticipants": [friend_3.relay_id, friend_4.relay_id],
            }
        },
    )
    content = response.json()
    assert content["data"]["chatRoomUpdate"]["room"]["node"]["title"] == "group"
    assert content["data"]["chatRoomUpdate"]["room"]["node"]["participantsCount"] == 3

    participants = content["data"]["chatRoomUpdate"]["room"]["node"]["participants"]["edges"]
    assert len(participants) == 3
    ids = [participant["node"]["profile"]["id"] for participant in participants]
    assert friend_1.relay_id in ids
    assert friend_2.relay_id in ids
    assert friend_3.relay_id not in ids
    assert friend_4.relay_id not in ids

    removed_participants = content["data"]["chatRoomUpdate"]["removedParticipants"]
    assert len(removed_participants) == 2
    ids = [participant["profile"]["id"] for participant in removed_participants]
    assert friend_1.relay_id not in ids
    assert friend_2.relay_id not in ids
    assert friend_3.relay_id in ids
    assert friend_4.relay_id in ids


def test_member_user_can_leave_room(django_user_client, graphql_user_client, django_client):
    friend_1 = ProfileFactory()
    friend_2 = ProfileFactory()
    friend_3 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend_1.relay_id,
                    friend_2.relay_id,
                    friend_3.relay_id,
                ],
            }
        },
    )
    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]

    django_client.force_login(friend_3.owner)

    response = graphql_query(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": friend_3.relay_id,
                "roomId": room_id,
                "removeParticipants": [friend_3.relay_id],
            }
        },
        client=django_client,
    )
    content = response.json()
    assert content["data"]["chatRoomUpdate"]["room"]["node"]["title"] == "group"
    participants_ids = [
        participant["node"]["profile"]["id"]
        for participant in content["data"]["chatRoomUpdate"]["room"]["node"]["participants"][
            "edges"
        ]
    ]
    assert len(participants_ids) == 3
    assert friend_3.relay_id not in participants_ids


def test_member_user_cannot_remove_other_members(
    django_user_client, graphql_user_client, django_client
):
    friend_1 = ProfileFactory()
    friend_2 = ProfileFactory()
    friend_3 = ProfileFactory()

    response = graphql_user_client(
        CREATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": django_user_client.user.profile.relay_id,
                "isGroup": True,
                "title": "group",
                "participants": [
                    friend_1.relay_id,
                    friend_2.relay_id,
                    friend_3.relay_id,
                ],
            }
        },
    )
    content = response.json()
    room_id = content["data"]["chatRoomCreate"]["room"]["node"]["id"]
    django_client.force_login(friend_3.owner)

    response = graphql_query(
        UPDATE_ROOM_GRAPHQL,
        variables={
            "input": {
                "profileId": friend_3.relay_id,
                "roomId": room_id,
                "removeParticipants": [friend_2.relay_id, friend_3.relay_id],
            }
        },
        client=django_client,
    )
    content = response.json()
    assert (
        content["data"]["chatRoomUpdate"]["errors"][0]["messages"][0]
        == "You don't have permission to update this room"
    )



# Generated by Django 4.2.15 on 2024-09-25 15:27

import baseapp_core.models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import swapper
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        swapper.dependency("baseapp_profiles", "Profile"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        swapper.dependency("baseapp_chats", "ChatRoom"),
        swapper.dependency("baseapp_chats", "ChatRoomParticipant"),
        swapper.dependency("baseapp_chats", "Message"),
        swapper.dependency("baseapp_chats", "MessageStatus"),
        swapper.dependency("baseapp_chats", "UnreadMessageCount"),
    ]

    operations = [
        migrations.CreateModel(
            name="ChatRoom",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="modified"
                    ),
                ),
                ("title", models.CharField(blank=True, max_length=255, null=True)),
                (
                    "image",
                    models.ImageField(
                        blank=True,
                        null=True,
                        upload_to=baseapp_core.models.random_name_in("chat_room_images"),
                        verbose_name="image",
                    ),
                ),
                (
                    "last_message_time",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now, null=True),
                ),
                ("participants_count", models.IntegerField(default=0)),
                ("messages_count", models.IntegerField(default=0)),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="created_rooms",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ["-last_message_time", "-created"],
                "abstract": False,
                "swappable": swapper.swappable_setting("baseapp_chats", "ChatRoom"),
            },
        ),
        migrations.CreateModel(
            name="Message",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="modified"
                    ),
                ),
                ("content", models.TextField(blank=True, null=True)),
                (
                    "verb",
                    models.IntegerField(
                        choices=[(100, "sent a message")], db_index=True, default=100
                    ),
                ),
                (
                    "action_object_object_id",
                    models.IntegerField(blank=True, db_index=True, null=True),
                ),
                ("extra_data", models.JSONField(blank=True, null=True)),
                (
                    "action_object_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="action_object",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "in_reply_to",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="replies",
                        to=swapper.get_model_name("baseapp_chats", "Message"),
                    ),
                ),
                (
                    "profile",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name("baseapp_profiles", "Profile"),
                    ),
                ),
                (
                    "room",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="messages",
                        to=swapper.get_model_name("baseapp_chats", "ChatRoom"),
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ["-created"],
                "abstract": False,
                "swappable": swapper.swappable_setting("baseapp_chats", "Message"),
            },
        ),
        migrations.CreateModel(
            name="ChatRoomParticipant",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="modified"
                    ),
                ),
                ("role", models.IntegerField(choices=[(1, "member"), (2, "admin")], default=1)),
                ("accepted_at", models.DateTimeField(blank=True, null=True)),
                (
                    "profile",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name("baseapp_profiles", "Profile"),
                    ),
                ),
                (
                    "room",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="participants",
                        to=swapper.get_model_name("baseapp_chats", "ChatRoom"),
                    ),
                ),
            ],
            options={
                "ordering": ["created"],
                "abstract": False,
                "swappable": swapper.swappable_setting("baseapp_chats", "ChatRoomParticipant"),
            },
        ),
        migrations.AddField(
            model_name="chatroom",
            name="last_message",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=swapper.get_model_name("baseapp_chats", "Message"),
            ),
        ),
        migrations.CreateModel(
            name="UnreadMessageCount",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("count", models.IntegerField(default=0)),
                (
                    "profile",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name("baseapp_profiles", "Profile"),
                    ),
                ),
                (
                    "room",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name("baseapp_chats", "ChatRoom"),
                    ),
                ),
            ],
            options={
                "ordering": ["id"],
                "abstract": False,
                "swappable": swapper.swappable_setting("baseapp_chats", "UnreadMessageCount"),
                "unique_together": {("room_id", "profile_id")},
            },
        ),
        migrations.CreateModel(
            name="MessageStatus",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("is_read", models.BooleanField(default=False)),
                ("read_at", models.DateTimeField(blank=True, null=True)),
                (
                    "message",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="statuses",
                        to=swapper.get_model_name("baseapp_chats", "Message"),
                    ),
                ),
                (
                    "profile",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name("baseapp_profiles", "Profile"),
                    ),
                ),
            ],
            options={
                "abstract": False,
                "swappable": swapper.swappable_setting("baseapp_chats", "MessageStatus"),
                "unique_together": {("message_id", "profile_id")},
            },
        ),
    ]



import graphene
import swapper
from baseapp_core.graphql import (
    RelayMutation,
    get_obj_from_relay_id,
    get_pk_from_relay_id,
    login_required,
)
from django.contrib.auth import get_user_model
from django.db import transaction
from django.db.models import Count, Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from graphene_django.types import ErrorType
from rest_framework import serializers

from baseapp_chats.graphql.subscriptions import (
    ChatRoomOnMessagesCountUpdate,
    ChatRoomOnNewMessage,
    ChatRoomOnRoomUpdate,
)
from baseapp_chats.utils import (
    CONTENT_LINKED_PROFILE_ACTOR,
    send_message,
    send_new_chat_message_notification,
)

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")
ChatRoomParticipantRoles = ChatRoomParticipant.ChatRoomParticipantRoles
Message = swapper.load_model("baseapp_chats", "Message")
MessageStatus = swapper.load_model("baseapp_chats", "MessageStatus")
UnreadMessageCount = swapper.load_model("baseapp_chats", "UnreadMessageCount")
Block = swapper.load_model("baseapp_blocks", "Block")
User = get_user_model()
Profile = swapper.load_model("baseapp_profiles", "Profile")


ChatRoomObjectType = ChatRoom.get_graphql_object_type()
ProfileObjectType = Profile.get_graphql_object_type()
MessageObjectType = Message.get_graphql_object_type()
ChatRoomParticipantObjectType = ChatRoomParticipant.get_graphql_object_type()


class ImageSerializer(serializers.Serializer):
    image = serializers.ImageField(required=False, allow_null=True)


class ChatRoomCreate(RelayMutation):
    room = graphene.Field(ChatRoomObjectType._meta.connection.Edge)
    profile = graphene.Field(ProfileObjectType)

    class Input:
        profile_id = graphene.ID(required=True)
        participants = graphene.List(graphene.ID, required=True)
        is_group = graphene.Boolean(required=False, default_value=False)
        title = graphene.String(required=False)

    @classmethod
    @login_required
    def mutate_and_get_payload(cls, root, info, profile_id, participants, is_group, **input):
        profile = get_obj_from_relay_id(info, profile_id)

        if not info.context.user.has_perm("baseapp_profiles.use_profile", profile):
            return ChatRoomCreate(
                errors=[
                    ErrorType(
                        field="profile_id",
                        messages=[_("You don't have permission to send a message as this profile")],
                    )
                ]
            )

        participants = [get_obj_from_relay_id(info, participant) for participant in participants]
        participants = [participant for participant in participants if participant is not None]
        participants_ids = [participant.pk for participant in participants]

        # Check if participants are blocked
        if Block.objects.filter(
            Q(actor_id=profile.id, target_id__in=participants_ids)
            | Q(actor_id__in=participants_ids, target_id=profile.id)
        ).exists():
            return ChatRoomCreate(
                errors=[
                    ErrorType(
                        field="participants",
                        messages=[_("You can't create a chatroom with those participants")],
                    )
                ]
            )

        participants.append(profile)

        if len(participants) < 2:
            return ChatRoomCreate(
                errors=[
                    ErrorType(
                        field="participants",
                        messages=[_("You need add at least one participant")],
                    )
                ]
            )

        is_group = is_group or len(participants) > 2
        title = input.get("title", None)
        if is_group and not title:
            return ChatRoomCreate(
                errors=[
                    ErrorType(
                        field="title",
                        messages=[_("Title is required for group chats")],
                    )
                ]
            )

        if not info.context.user.has_perm(
            "baseapp_chats.add_chatroom", {"profile": profile, "participants": participants}
        ):
            return ChatRoomCreate(
                errors=[
                    ErrorType(
                        field="participants",
                        messages=[_("You don't have permission to create a room")],
                    )
                ]
            )

        query_set = ChatRoom.objects.annotate(count=Count("participants")).filter(
            count=len(participants)
        )
        for participant in participants:
            query_set = query_set.filter(
                participants__profile_id=participant.id,
                is_group=False,
            )
        existent_room = query_set.first()

        if existent_room and not is_group:
            return ChatRoomCreate(
                profile=profile,
                room=ChatRoomObjectType._meta.connection.Edge(
                    node=existent_room,
                ),
            )
        image = info.context.FILES.get("image", None)
        serializer = ImageSerializer(data={"image": image})
        if not serializer.is_valid():
            return ChatRoomCreate(
                errors=[ErrorType(field="image", messages=serializer.errors["image"])]
            )

        room = ChatRoom.objects.create(
            created_by=info.context.user,
            last_message_time=timezone.now(),
            is_group=is_group,
            title=title,
            image=serializer.validated_data["image"],
        )

        created_participants = ChatRoomParticipant.objects.bulk_create(
            [
                ChatRoomParticipant(
                    profile=participant,
                    room=room,
                    role=(
                        ChatRoomParticipantRoles.ADMIN
                        if participant == profile
                        else ChatRoomParticipantRoles.MEMBER
                    ),
                    accepted_at=timezone.now(),
                )
                for participant in participants
            ]
        )

        room.participants_count = len(created_participants)
        room.save()

        if is_group:
            send_message(
                message_type=Message.MessageType.SYSTEM_GENERATED,
                room=room,
                profile=None,
                user=None,
                content=CONTENT_LINKED_PROFILE_ACTOR + ' created group "' + title + '"',
                content_linked_profile_actor=profile,
            )
            ChatRoomOnRoomUpdate.room_updated(room)

        return ChatRoomCreate(
            profile=profile,
            room=ChatRoomObjectType._meta.connection.Edge(
                node=room,
            ),
        )


class ChatRoomUpdate(RelayMutation):
    room = graphene.Field(ChatRoomObjectType._meta.connection.Edge)
    removed_participants = graphene.List(ChatRoomParticipantObjectType)

    class Input:
        room_id = graphene.ID(required=True)
        profile_id = graphene.ID(required=True)
        title = graphene.String(required=False)
        delete_image = graphene.Boolean(default_value=False)
        add_participants = graphene.List(graphene.ID, default_value=[])
        remove_participants = graphene.List(graphene.ID, default_value=[])

    @classmethod
    @login_required
    def mutate_and_get_payload(
        cls,
        root,
        info,
        room_id,
        profile_id,
        delete_image,
        add_participants,
        remove_participants,
        **input,
    ):
        room = get_obj_from_relay_id(info, room_id)
        profile = get_obj_from_relay_id(info, profile_id)
        is_sole_admin = (
            ChatRoomParticipant.objects.filter(
                profile=profile, room=room, role=ChatRoomParticipantRoles.ADMIN
            ).exists()
            and ChatRoomParticipant.objects.filter(
                room=room, role=ChatRoomParticipantRoles.ADMIN
            ).count()
            == 1
        )
        is_leaving_chatroom = profile_id in remove_participants and len(remove_participants) == 1

        add_participants_pks = [
            get_pk_from_relay_id(participant) for participant in add_participants
        ]

        remove_participants_pks = [
            get_pk_from_relay_id(participant) for participant in remove_participants
        ]
        participants_to_remove = ChatRoomParticipant.objects.filter(
            profile_id__in=remove_participants_pks, room=room
        )
        removed_participants = list(participants_to_remove)

        title = input.get("title", None)
        image = info.context.FILES.get("image", None)

        if not room or not getattr(room, "is_group", False):
            return ChatRoomUpdate(
                errors=[
                    ErrorType(
                        field="room_id",
                        messages=[_("This room cannot be updated")],
                    )
                ]
            )

        if not info.context.user.has_perm("baseapp_profiles.use_profile", profile):
            return ChatRoomUpdate(
                errors=[
                    ErrorType(
                        field="profile_id",
                        messages=[_("You don't have permission to act as this profile")],
                    )
                ]
            )

        # Check if added participants are blocked
        if Block.objects.filter(
            Q(actor_id=profile.id, target_id__in=add_participants_pks)
            | Q(actor_id__in=add_participants_pks, target_id=profile.id)
        ).exists():
            return ChatRoomUpdate(
                errors=[
                    ErrorType(
                        field="add_participants",
                        messages=[_("You can't add those participants to a chatroom")],
                    )
                ]
            )

        if not info.context.user.has_perm(
            "baseapp_chats.modify_chatroom",
            {
                "profile": profile,
                "room": room,
                "add_participants": add_participants_pks,
                "is_leaving_chatroom": is_leaving_chatroom,
                "modify_image": image or delete_image,
                "modify_title": title,
            },
        ):
            return ChatRoomUpdate(
                errors=[
                    ErrorType(
                        field="room_id",
                        messages=[_("You don't have permission to update this room")],
                    )
                ]
            )

        # Change image
        serializer = ImageSerializer(data={"image": image})
        if not serializer.is_valid():
            return ChatRoomUpdate(
                errors=[ErrorType(field="image", messages=serializer.errors["image"])]
            )

        participants_to_remove.delete()

        # Setting new admin if needed
        if is_leaving_chatroom and is_sole_admin:
            oldest_remaining_participant = (
                ChatRoomParticipant.objects.filter(room=room).order_by("accepted_at").first()
            )
            if oldest_remaining_participant:
                oldest_remaining_participant.role = ChatRoomParticipantRoles.ADMIN
                oldest_remaining_participant.save(update_fields=["role"])

        # Update room
        # TODO: Remeber to include added participants into count when implementing that feature
        room.participants_count = room.participants_count - len(removed_participants)
        if title is not None:
            room.title = title
        if image is not None:
            room.image = serializer.validated_data["image"]
        elif delete_image:
            room.image = None
        room.save()

        ChatRoomOnRoomUpdate.room_updated(room, removed_participants)

        return ChatRoomUpdate(
            room=ChatRoomObjectType._meta.connection.Edge(
                node=room,
            ),
            removed_participants=removed_participants,
        )


class ChatRoomSendMessage(RelayMutation):
    message = graphene.Field(MessageObjectType._meta.connection.Edge)

    class Input:
        room_id = graphene.ID(required=True)
        profile_id = graphene.ID(required=True)
        content = graphene.String(required=True)
        in_reply_to_id = graphene.ID(required=False)

    @classmethod
    @login_required
    def mutate_and_get_payload(
        cls, root, info, room_id, content, profile_id, in_reply_to_id=None, **input
    ):
        room = get_obj_from_relay_id(info, room_id)
        profile = get_obj_from_relay_id(info, profile_id)

        in_reply_to = None
        if in_reply_to_id:
            in_reply_to_pk = get_pk_from_relay_id(in_reply_to_id)
            in_reply_to = Message.objects.filter(
                pk=in_reply_to_pk,
                room=room,
            ).first()

        if not info.context.user.has_perm("baseapp_profiles.use_profile", profile):
            return ChatRoomSendMessage(
                errors=[
                    ErrorType(
                        field="profile_id",
                        messages=[_("You don't have permission to send a message as this profile")],
                    )
                ]
            )

        if not room or not info.context.user.has_perm(
            "baseapp_chats.add_message", {"profile": profile, "room": room}
        ):
            return ChatRoomSendMessage(
                errors=[
                    ErrorType(
                        field="room_id",
                        messages=[_("You don't have permission to send a message in this room")],
                    )
                ]
            )

        if len(content) < 1:
            return ChatRoomSendMessage(
                errors=[
                    ErrorType(
                        field="content",
                        messages=[_("You need to write something")],
                    )
                ]
            )

        if len(content) > 1000:
            return ChatRoomSendMessage(
                errors=[
                    ErrorType(
                        field="content",
                        messages=[_("You can't write more than 1000 characters")],
                    )
                ]
            )

        message = send_message(
            profile=profile,
            user=info.context.user,
            content=content,
            room=room,
            verb=Message.Verbs.SENT_MESSAGE,
            room_id=room_id,
            in_reply_to=in_reply_to,
        )

        send_new_chat_message_notification(room, message, info)
        ChatRoomReadMessages.read_messages(room, profile)

        return ChatRoomSendMessage(
            message=MessageObjectType._meta.connection.Edge(
                node=message,
            )
        )


class ChatRoomEditMessage(RelayMutation):
    message = graphene.Field(MessageObjectType._meta.connection.Edge)

    class Input:
        id = graphene.ID(required=True)
        content = graphene.String(required=True)

    @classmethod
    @login_required
    def mutate_and_get_payload(cls, root, info, **input):
        pk = get_pk_from_relay_id(input.get("id"))
        try:
            message = Message.objects.get(pk=pk)
        except Message.DoesNotExist:
            return ChatRoomEditMessage(
                errors=[
                    ErrorType(
                        field="id",
                        messages=[_("Message does not exist")],
                    )
                ]
            )

        profile = (
            info.context.user.current_profile
            if hasattr(info.context.user, "current_profile")
            else (info.context.user.profile if hasattr(info.context.user, "profile") else None)
        )
        if not info.context.user.has_perm(
            "baseapp_chats.change_message",
            {
                "profile": profile,
                "message": message,
            },
        ):
            return ChatRoomEditMessage(
                errors=[
                    ErrorType(
                        field="id",
                        messages=[_("You don't have permission to update this message")],
                    )
                ]
            )

        content = input.get("content")
        if len(content) < 1:
            return ChatRoomEditMessage(
                errors=[
                    ErrorType(
                        field="content",
                        messages=[_("You cannot edit an empty message")],
                    )
                ]
            )

        if len(content) > 1000:
            return ChatRoomEditMessage(
                errors=[
                    ErrorType(
                        field="content",
                        messages=[_("Message must be no longer than 1000 characters")],
                    )
                ]
            )

        message.content = content
        message.save(update_fields=["content"])

        ChatRoomOnNewMessage.new_message(room_id=message.room.relay_id, message=message)

        return ChatRoomEditMessage(
            message=MessageObjectType._meta.connection.Edge(
                node=message,
            )
        )


class ChatRoomReadMessages(RelayMutation):
    room = graphene.Field(ChatRoomObjectType)
    profile = graphene.Field(ProfileObjectType)
    messages = graphene.List(MessageObjectType)

    class Input:
        room_id = graphene.ID(required=True)
        profile_id = graphene.ID(required=True)
        message_ids = graphene.List(graphene.ID, required=False)

    @classmethod
    @login_required
    def mutate_and_get_payload(cls, root, info, room_id, profile_id, message_ids=None, **input):
        room = get_obj_from_relay_id(info, room_id)
        profile = get_obj_from_relay_id(info, profile_id)

        if not info.context.user.has_perm("baseapp_profiles.use_profile", profile):
            return ChatRoomReadMessages(
                errors=[
                    ErrorType(
                        field="profile_id",
                        messages=[_("You don't have permission to act as this profile")],
                    )
                ]
            )

        if not ChatRoomParticipant.objects.filter(
            profile_id=profile.pk,
            room=room,
        ).exists():
            return ChatRoomReadMessages(
                errors=[
                    ErrorType(
                        field="participant",
                        messages=[_("Participant is not part of the room.")],
                    )
                ]
            )

        cls.remove_marked_unread(room, profile)
        return cls.read_messages(room, profile, message_ids)

    @classmethod
    def remove_marked_unread(cls, room, profile):
        UnreadMessageCount.objects.filter(profile=profile, room=room, marked_unread=True).update(
            marked_unread=False
        )

    @classmethod
    def read_messages(cls, room, profile, message_ids=None):
        messages_status_qs = MessageStatus.objects.filter(
            profile_id=profile.pk,
            is_read=False,
        )

        if message_ids:
            message_ids = [get_pk_from_relay_id(message_id) for message_id in message_ids]
            messages_status_qs = messages_status_qs.filter(
                message_id__in=message_ids,
            )
        else:
            messages_status_qs = messages_status_qs.filter(
                message__room_id=room.pk,
            )

        messages = Message.objects.filter(
            pk__in=messages_status_qs.values_list("message_id", flat=True)
        )

        messages_status_qs.update(is_read=True, read_at=timezone.now())

        ChatRoomOnMessagesCountUpdate.send_updated_chat_count(
            profile=profile, profile_id=profile.relay_id
        )

        return ChatRoomReadMessages(room=room, profile=profile, messages=messages)


class ChatRoomUnread(RelayMutation):
    room = graphene.Field(ChatRoomObjectType)
    profile = graphene.Field(ProfileObjectType)

    class Input:
        room_id = graphene.ID(required=True)
        profile_id = graphene.ID(required=True)

    @classmethod
    @login_required
    def mutate_and_get_payload(cls, root, info, room_id, profile_id, **input):
        room = get_obj_from_relay_id(info, room_id)
        profile = get_obj_from_relay_id(info, profile_id)

        if not info.context.user.has_perm("baseapp_profiles.use_profile", profile):
            return ChatRoomUnread(
                errors=[
                    ErrorType(
                        field="profile_id",
                        messages=[_("You don't have permission to act as this profile")],
                    )
                ]
            )

        if not ChatRoomParticipant.objects.filter(
            profile_id=profile.pk,
            room=room,
        ).exists():
            return ChatRoomUnread(
                errors=[
                    ErrorType(
                        field="participant",
                        messages=[_("Participant is not part of the room.")],
                    )
                ]
            )

        UnreadMessageCount.objects.update_or_create(
            profile=profile,
            room=room,
            defaults={"marked_unread": True},
        )

        return ChatRoomUnread(room=room, profile=profile)


class ChatRoomArchive(RelayMutation):
    room = graphene.Field(ChatRoomObjectType)

    class Input:
        room_id = graphene.ID(required=True)
        profile_id = graphene.ID(required=True)
        archive = graphene.Boolean(required=True)

    @classmethod
    @login_required
    def mutate_and_get_payload(cls, root, info, room_id, profile_id, archive, **input):
        room = get_obj_from_relay_id(info, room_id)
        profile = get_obj_from_relay_id(info, profile_id)

        if not info.context.user.has_perm("baseapp_profiles.use_profile", profile):
            return ChatRoomArchive(
                errors=[
                    ErrorType(
                        field="profile_id",
                        messages=[
                            _("You don't have permission to archive this chatroom as this profile")
                        ],
                    )
                ]
            )

        participant = ChatRoomParticipant.objects.filter(
            profile_id=profile.pk,
            room=room,
        ).first()

        if not participant:
            return ChatRoomArchive(
                errors=[
                    ErrorType(
                        field="participant",
                        messages=[_("Participant is not part of the room.")],
                    )
                ]
            )

        with transaction.atomic():
            participant = ChatRoomParticipant.objects.select_for_update().get(pk=participant.pk)
            participant.has_archived_room = archive
            participant.save()
            room.refresh_from_db()
        return ChatRoomArchive(room=room)


class ChatsMutations(object):
    chat_room_create = ChatRoomCreate.Field()
    chat_room_update = ChatRoomUpdate.Field()
    chat_room_send_message = ChatRoomSendMessage.Field()
    chat_room_edit_message = ChatRoomEditMessage.Field()
    chat_room_read_messages = ChatRoomReadMessages.Field()
    chat_room_unread = ChatRoomUnread.Field()
    chat_room_archive = ChatRoomArchive.Field()



from datetime import timedelta

import pytest
import swapper
from baseapp_blocks.tests.factories import BlockFactory
from baseapp_core.graphql.testing.fixtures import graphql_query
from baseapp_core.tests.factories import UserFactory
from baseapp_profiles.tests.factories import ProfileFactory
from django.utils import timezone
from freezegun import freeze_time

from .factories import ChatRoomFactory, ChatRoomParticipantFactory, MessageFactory

pytestmark = pytest.mark.django_db

UnreadMessageCount = swapper.load_model("baseapp_chats", "UnreadMessageCount")


USER_ROOMS_GRAPHQL = """
    query UserRooms {
        me {
            profile {
                chatRooms {
                    edges {
                        node {
                            id
                        }
                    }
                }
            }
        }
    }
"""

ROOM_GRAPHQL = """
    query GetRoom($roomId: ID!) {
        chatRoom(id: $roomId) {
            allMessages {
                edges {
                    node {
                        id
                        content
                        messageType
                    }
                }
            }
        }
    }
"""

PROFILE_ROOMS_GRAPHQL = """
    query ProfileRooms($profileId: ID!, $archived: Boolean, $unreadMessages: Boolean, $q: String) {
        profile(id: $profileId) {
            id
            name
            ... on ChatRoomsInterface {
                chatRooms (archived: $archived, unreadMessages: $unreadMessages, q: $q) {
                    edges {
                        node {
                            id
                            unreadMessages(profileId: $profileId) {
                                count
                                markedUnread
                            }
                        }
                    }
                }
            }
        }
    }
"""


def test_user_can_list_rooms(graphql_user_client, django_user_client):
    user_profie = ProfileFactory()
    friend_profile = ProfileFactory()

    user_room = ChatRoomFactory(created_by=user_profie.owner)
    friend_room = ChatRoomFactory(created_by=friend_profile.owner)

    ChatRoomParticipantFactory(profile=user_profie, room=user_room)
    ChatRoomParticipantFactory(profile=friend_profile, room=friend_room)

    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=user_room)
    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=friend_room)

    response = graphql_user_client(
        USER_ROOMS_GRAPHQL,
        variables={},
    )

    content = response.json()

    assert len(content["data"]["me"]["profile"]["chatRooms"]["edges"]) == 2


def test_unread_messages_count(graphql_user_client, django_user_client):
    my_profile = django_user_client.user.profile
    friend_profile = ProfileFactory()
    room = ChatRoomFactory(created_by=django_user_client.user)

    ChatRoomParticipantFactory(room=room, profile=my_profile)
    ChatRoomParticipantFactory(room=room, profile=friend_profile)

    MessageFactory(room=room, profile=friend_profile)

    response = graphql_user_client(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": my_profile.relay_id},
    )

    content = response.json()

    assert len(content["data"]["profile"]["chatRooms"]["edges"]) == 1
    assert (
        content["data"]["profile"]["chatRooms"]["edges"][0]["node"]["unreadMessages"]["count"] == 1
    )


def test_filter_rooms_with_unread_messages(graphql_user_client, django_user_client):
    my_profile = django_user_client.user.profile
    friend_in_room_with_unread = ProfileFactory()
    friend_in_room_with_read = ProfileFactory()
    room_with_unread = ChatRoomFactory(created_by=django_user_client.user)
    room_with_read = ChatRoomFactory(created_by=django_user_client.user)

    ChatRoomParticipantFactory(room=room_with_unread, profile=my_profile)
    ChatRoomParticipantFactory(room=room_with_unread, profile=friend_in_room_with_unread)

    MessageFactory(room=room_with_unread, profile=friend_in_room_with_unread)

    ChatRoomParticipantFactory(room=room_with_read, profile=my_profile)
    ChatRoomParticipantFactory(room=room_with_read, profile=friend_in_room_with_read)

    read_message = MessageFactory(room=room_with_read, profile=friend_in_room_with_read)

    read_message.statuses.filter(profile=my_profile).update(is_read=True)

    response = graphql_user_client(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": my_profile.relay_id, "unreadMessages": True},
    )

    content = response.json()

    assert len(content["data"]["profile"]["chatRooms"]["edges"]) == 1
    assert (
        content["data"]["profile"]["chatRooms"]["edges"][0]["node"]["id"]
        == room_with_unread.relay_id
    )


def test_archived_chats(django_client):
    me = UserFactory()
    my_profile = me.profile
    friend = UserFactory()
    friend_profile = friend.profile
    room = ChatRoomFactory(created_by=me)

    ChatRoomParticipantFactory(room=room, profile=my_profile, has_archived_room=True)
    ChatRoomParticipantFactory(room=room, profile=friend_profile)

    django_client.force_login(me)
    response = graphql_query(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": my_profile.relay_id, "archived": True},
        client=django_client,
    )

    content = response.json()

    assert len(content["data"]["profile"]["chatRooms"]["edges"]) == 1

    django_client.force_login(friend)
    response = graphql_query(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": friend_profile.relay_id, "archived": True},
        client=django_client,
    )

    content = response.json()

    assert len(content["data"]["profile"]["chatRooms"]["edges"]) == 0


def test_can_filter_by_room_title(django_client):
    user_1 = UserFactory(first_name="Luke", last_name="Skywalker")
    profile_1 = user_1.profile

    user_2 = UserFactory(first_name="Darth", last_name="Vader")
    profile_2 = user_2.profile

    profile_3 = ProfileFactory(name="Tester123")
    profile_4 = ProfileFactory(name="Luke Littler")

    room_1 = ChatRoomFactory(created_by=profile_1.owner, is_group=False)
    room_2 = ChatRoomFactory(created_by=profile_1.owner, is_group=False)
    room_3 = ChatRoomFactory(created_by=profile_1.owner, title="Lucky Luke", is_group=True)
    room_4 = ChatRoomFactory(created_by=profile_2.owner, title="Star Wars", is_group=True)

    ChatRoomParticipantFactory(room=room_1, profile=profile_1)
    ChatRoomParticipantFactory(room=room_1, profile=profile_2, has_archived_room=True)

    ChatRoomParticipantFactory(room=room_2, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_2, profile=profile_4)

    ChatRoomParticipantFactory(room=room_3, profile=profile_1)
    ChatRoomParticipantFactory(room=room_3, profile=profile_2)
    ChatRoomParticipantFactory(room=room_3, profile=profile_3, has_archived_room=True)

    ChatRoomParticipantFactory(room=room_4, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_4, profile=profile_2)

    django_client.force_login(user_1)
    response = graphql_query(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": profile_1.relay_id, "q": "lUKE"},
        client=django_client,
    )
    content = response.json()

    rooms = content["data"]["profile"]["chatRooms"]["edges"]
    assert len(rooms) == 2
    room_ids = [room["node"]["id"] for room in rooms]
    assert room_1.relay_id not in room_ids  # room name 'Darth Vader'
    assert room_2.relay_id in room_ids  # room name 'Luke Littler'
    assert room_3.relay_id in room_ids  # room name 'Lucky Luke'
    assert room_4.relay_id not in room_ids  # room name 'Star Wars'

    django_client.force_login(user_2)
    response = graphql_query(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": profile_2.relay_id, "q": "u"},
        client=django_client,
    )
    content = response.json()

    rooms = content["data"]["profile"]["chatRooms"]["edges"]
    assert len(rooms) == 2
    room_ids = [room["node"]["id"] for room in rooms]
    assert room_1.relay_id in room_ids  # room name 'Luke Skywalker'
    assert room_2.relay_id not in room_ids  # not a member
    assert room_3.relay_id in room_ids  # room name 'Lucky Luke'
    assert room_4.relay_id not in room_ids  # room name 'Star Wars'


def test_can_filter_unread_rooms_by_title(django_client):
    user_1 = UserFactory(first_name="Luke", last_name="Skywalker (Artist)")
    profile_1 = user_1.profile
    profile_2 = ProfileFactory(name="Darth Vader")
    profile_3 = ProfileFactory(name="Han Solo")
    profile_4 = ProfileFactory(name="R2D2")

    room_1 = ChatRoomFactory(created_by=profile_2.owner, is_group=False)
    room_2 = ChatRoomFactory(created_by=profile_3.owner, is_group=False)
    room_3 = ChatRoomFactory(created_by=profile_4.owner, title="art exhibition", is_group=True)
    room_4 = ChatRoomFactory(created_by=profile_1.owner, title="Some group", is_group=True)
    room_5 = ChatRoomFactory(created_by=profile_1.owner, title="Star Wars, Part I", is_group=True)

    ChatRoomParticipantFactory(room=room_1, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_1, profile=profile_2)

    ChatRoomParticipantFactory(room=room_2, profile=profile_1)
    ChatRoomParticipantFactory(room=room_2, profile=profile_4, has_archived_room=True)

    ChatRoomParticipantFactory(room=room_3, profile=profile_1)
    ChatRoomParticipantFactory(room=room_3, profile=profile_2, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_3, profile=profile_3)

    ChatRoomParticipantFactory(room=room_4, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_4, profile=profile_2)

    ChatRoomParticipantFactory(room=room_5, profile=profile_1)
    ChatRoomParticipantFactory(room=room_5, profile=profile_4)

    UnreadMessageCount(room=room_1, profile=profile_1, marked_unread=True).save()
    MessageFactory(room=room_2, profile=profile_4)
    MessageFactory(room=room_3, profile=profile_3)
    MessageFactory(room=room_4, profile=profile_2)
    MessageFactory(room=room_5, profile=profile_1)

    django_client.force_login(user_1)
    response = graphql_query(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": profile_1.relay_id, "unreadMessages": True, "q": "art"},
        client=django_client,
    )
    content = response.json()

    rooms = content["data"]["profile"]["chatRooms"]["edges"]
    assert len(rooms) == 2
    room_ids = [room["node"]["id"] for room in rooms]
    assert room_1.relay_id in room_ids  # room name 'Darth Vader'
    assert room_2.relay_id not in room_ids  # room name 'R2D2'
    assert room_3.relay_id in room_ids  # room name 'art exhibition'
    assert room_4.relay_id not in room_ids  # room name 'Some group'
    assert room_5.relay_id not in room_ids  # room name 'Star Wars, Part I', but no unread messages


def test_can_filter_archived_rooms_by_title(django_client):
    user_1 = UserFactory(first_name="Tester", last_name="11")
    profile_1 = user_1.profile
    profile_2 = ProfileFactory(name="Shrek")
    profile_3 = ProfileFactory(name="111111")
    profile_4 = ProfileFactory(name="112358")

    room_1 = ChatRoomFactory(created_by=profile_4.owner, title="11th street", is_group=True)
    room_2 = ChatRoomFactory(created_by=profile_4.owner, is_group=False)
    room_3 = ChatRoomFactory(created_by=profile_4.owner, title="10000 BC", is_group=True)
    room_4 = ChatRoomFactory(created_by=profile_4.owner, is_group=False)
    room_5 = ChatRoomFactory(created_by=profile_4.owner, title="Oceans 11", is_group=True)

    ChatRoomParticipantFactory(room=room_1, profile=profile_1)
    ChatRoomParticipantFactory(room=room_1, profile=profile_2, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_1, profile=profile_4, has_archived_room=True)

    ChatRoomParticipantFactory(room=room_2, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_2, profile=profile_2, has_archived_room=True)

    ChatRoomParticipantFactory(room=room_3, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_3, profile=profile_2, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_3, profile=profile_3, has_archived_room=True)

    ChatRoomParticipantFactory(room=room_4, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_4, profile=profile_3)

    ChatRoomParticipantFactory(room=room_5, profile=profile_1, has_archived_room=True)
    ChatRoomParticipantFactory(room=room_5, profile=profile_3)

    MessageFactory(room=room_2, profile=profile_4)
    MessageFactory(room=room_5, profile=profile_3)

    django_client.force_login(user_1)
    response = graphql_query(
        PROFILE_ROOMS_GRAPHQL,
        variables={"profileId": profile_1.relay_id, "archived": True, "q": "11"},
        client=django_client,
    )
    content = response.json()

    rooms = content["data"]["profile"]["chatRooms"]["edges"]
    assert len(rooms) == 2
    room_ids = [room["node"]["id"] for room in rooms]
    assert room_1.relay_id not in room_ids  # room name '11th street' but not archived
    assert room_2.relay_id not in room_ids  # room name 'Shrek'
    assert room_3.relay_id not in room_ids  # room name '10000 BC'
    assert room_4.relay_id in room_ids  # room name '112358'
    assert room_5.relay_id in room_ids  # room name 'Oceans 11'


def test_cant_list_rooms_if_not_participating(graphql_user_client):
    user = UserFactory()

    user_room = ChatRoomFactory(created_by=user)

    ChatRoomParticipantFactory(profile=user.profile, room=user_room)

    response = graphql_user_client(
        USER_ROOMS_GRAPHQL,
        variables={},
    )

    content = response.json()

    assert len(content["data"]["me"]["profile"]["chatRooms"]["edges"]) == 0


def test_rooms_list_are_ordered_by_last_message_time(graphql_user_client, django_user_client):
    user = UserFactory()
    user_2 = UserFactory()
    user_3 = UserFactory()

    room = ChatRoomFactory(created_by=user)
    room_2 = ChatRoomFactory(created_by=user_2, last_message_time=timezone.now())
    room_3 = ChatRoomFactory(
        created_by=user_3, last_message_time=timezone.now() - timedelta(hours=1)
    )

    ChatRoomParticipantFactory(profile=user.profile, room=room)
    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    ChatRoomParticipantFactory(profile=user_2.profile, room=room_2)
    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room_2)
    ChatRoomParticipantFactory(profile=user_3.profile, room=room_3)
    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room_3)

    room_id = room.relay_id
    room_2_id = room_2.relay_id
    room_3_id = room_3.relay_id

    response = graphql_user_client(
        USER_ROOMS_GRAPHQL,
        variables={},
    )

    content = response.json()

    assert len(content["data"]["me"]["profile"]["chatRooms"]["edges"]) == 3
    assert content["data"]["me"]["profile"]["chatRooms"]["edges"][0]["node"]["id"] == room_2_id
    assert content["data"]["me"]["profile"]["chatRooms"]["edges"][1]["node"]["id"] == room_id
    assert content["data"]["me"]["profile"]["chatRooms"]["edges"][2]["node"]["id"] == room_3_id


def test_can_list_messages_from_participating_room(graphql_user_client, django_user_client):
    band = ProfileFactory()
    room = ChatRoomFactory(created_by=band.owner)

    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    ChatRoomParticipantFactory(profile=band, room=room)

    message_1 = MessageFactory(room=room, profile=band)
    message_2 = MessageFactory(room=room, profile=band)

    response = graphql_user_client(ROOM_GRAPHQL, variables={"roomId": room.relay_id})

    content = response.json()

    assert len(content["data"]["chatRoom"]["allMessages"]["edges"]) == 2
    assert (
        content["data"]["chatRoom"]["allMessages"]["edges"][0]["node"]["content"]
        == message_1.content
    )
    assert (
        content["data"]["chatRoom"]["allMessages"]["edges"][1]["node"]["content"]
        == message_2.content
    )


def test_cant_list_messages_from_non_participating_room(graphql_user_client):
    user = UserFactory()
    room = ChatRoomFactory(created_by=user)

    ChatRoomParticipantFactory(profile=user.profile, room=room)

    MessageFactory(room=room, profile=user.profile)
    MessageFactory(room=room, profile=user.profile)

    response = graphql_user_client(ROOM_GRAPHQL, variables={"roomId": room.relay_id})

    content = response.json()
    assert content["data"]["chatRoom"] is None


def test_user_cant_list_rooms_if_blocked(graphql_user_client, django_user_client):
    band = ProfileFactory()
    room = ChatRoomFactory(created_by=django_user_client.user)

    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    ChatRoomParticipantFactory(profile=band, room=room)

    response = graphql_user_client(
        USER_ROOMS_GRAPHQL,
        variables={},
    )
    content = response.json()
    assert len(content["data"]["me"]["profile"]["chatRooms"]["edges"]) == 1

    BlockFactory(actor=band, target=django_user_client.user.profile)

    response = graphql_user_client(
        USER_ROOMS_GRAPHQL,
        variables={},
    )
    content = response.json()
    assert len(content["data"]["me"]["profile"]["chatRooms"]["edges"]) == 0


def test_user_can_list_rooms_that_are_not_blocked(graphql_user_client, django_user_client):
    band = ProfileFactory()
    user_2 = UserFactory()

    room = ChatRoomFactory(created_by=django_user_client.user)
    room_2 = ChatRoomFactory(created_by=user_2, last_message_time=timezone.now())

    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    ChatRoomParticipantFactory(profile=band, room=room)

    ChatRoomParticipantFactory(profile=user_2.profile, room=room_2)
    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room_2)

    response = graphql_user_client(
        USER_ROOMS_GRAPHQL,
        variables={},
    )
    content = response.json()
    assert len(content["data"]["me"]["profile"]["chatRooms"]["edges"]) == 2

    BlockFactory(actor=django_user_client.user.profile, target=band)

    response = graphql_user_client(
        USER_ROOMS_GRAPHQL,
        variables={},
    )
    content = response.json()
    assert len(content["data"]["me"]["profile"]["chatRooms"]["edges"]) == 1


def test_user_cant_open_room_if_blocked(graphql_user_client, django_user_client):
    band = ProfileFactory()
    room = ChatRoomFactory(created_by=django_user_client.user)

    ChatRoomParticipantFactory(profile=django_user_client.user.profile, room=room)
    ChatRoomParticipantFactory(profile=band, room=room)

    response = graphql_user_client(ROOM_GRAPHQL, variables={"roomId": room.relay_id})
    content = response.json()

    assert content["data"]["chatRoom"] is not None

    BlockFactory(actor=band, target=django_user_client.user.profile)

    response = graphql_user_client(ROOM_GRAPHQL, variables={"roomId": room.relay_id})
    content = response.json()
    assert content["data"]["chatRoom"] is None


def test_new_participant_cant_list_previous_messages_when_joining_group_room(
    graphql_user_client, django_user_client, django_client
):
    with freeze_time("2024-12-01 10:00:00") as frozen_time:
        existing_profile_1 = ProfileFactory()
        existing_profile_2 = ProfileFactory()
        room = ChatRoomFactory(created_by=existing_profile_1.owner, participants_count=2)

        ChatRoomParticipantFactory(
            profile=existing_profile_1, room=room, accepted_at=timezone.now()
        )
        ChatRoomParticipantFactory(
            profile=existing_profile_2, room=room, accepted_at=timezone.now()
        )

        MessageFactory(room=room, profile=existing_profile_1)
        MessageFactory(room=room, profile=existing_profile_2)

        frozen_time.move_to(timezone.now() + timedelta(hours=1, minutes=5))

        ChatRoomParticipantFactory(
            profile=django_user_client.user.profile, room=room, accepted_at=timezone.now()
        )
        room.participants_count = 3
        room.is_group = True
        room.save()
        room.refresh_from_db()

        MessageFactory(room=room, profile=existing_profile_1)
        MessageFactory(room=room, profile=existing_profile_2)

        response = graphql_user_client(ROOM_GRAPHQL, variables={"roomId": room.relay_id})

        content = response.json()

        assert len(content["data"]["chatRoom"]["allMessages"]["edges"]) == 2

        django_client.force_login(existing_profile_1.owner)

        response = graphql_query(
            ROOM_GRAPHQL, variables={"roomId": room.relay_id}, client=django_client
        )

        content = response.json()

        assert len(content["data"]["chatRoom"]["allMessages"]["edges"]) == 4



# Generated by Django 4.2.15 on 2024-09-25 15:29

import pgtrigger.compiler
import pgtrigger.migrations
import swapper
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("baseapp_chats", "0001_initial"),
    ]

    operations = filter(
        lambda x: x is not None,
        [
            (
                pgtrigger.migrations.AddTrigger(
                    model_name="message",
                    trigger=pgtrigger.compiler.Trigger(
                        name="set_last_message",
                        sql=pgtrigger.compiler.UpsertTriggerSql(
                            func="\n            UPDATE baseapp_chats_chatroom\n            SET last_message_id = NEW.id, last_message_time = NEW.created\n            WHERE id = NEW.room_id;\n            RETURN NULL;\n        ",
                            hash="06f966e2a58859020db8c4fe063d60775823bc1a",
                            operation="INSERT",
                            pgid="pgtrigger_set_last_message_9c677",
                            table="baseapp_chats_message",
                            when="AFTER",
                        ),
                    ),
                )
                if not swapper.is_swapped("baseapp_chats", "Message")
                else None
            ),
            (
                pgtrigger.migrations.AddTrigger(
                    model_name="message",
                    trigger=pgtrigger.compiler.Trigger(
                        name="create_message_status",
                        sql=pgtrigger.compiler.UpsertTriggerSql(
                            func="\n            INSERT INTO baseapp_chats_messagestatus (message_id, profile_id, is_read)\n            SELECT NEW.id, crp.profile_id, CASE WHEN NEW.profile_id = crp.profile_id THEN TRUE ELSE FALSE END\n            FROM baseapp_chats_chatroomparticipant as crp\n            WHERE crp.room_id = NEW.room_id;\n            RETURN NULL;\n        ",
                            hash="7a38e6c43d86b779e659c95fe1b790016ad3ca54",
                            operation="INSERT",
                            pgid="pgtrigger_create_message_status_c941b",
                            table="baseapp_chats_message",
                            when="AFTER",
                        ),
                    ),
                )
                if not swapper.is_swapped("baseapp_chats", "Message")
                else None
            ),
            (
                pgtrigger.migrations.AddTrigger(
                    model_name="message",
                    trigger=pgtrigger.compiler.Trigger(
                        name="update_last_message",
                        sql=pgtrigger.compiler.UpsertTriggerSql(
                            func="\n            UPDATE baseapp_chats_chatroom\n            SET last_message_id = (\n                SELECT id\n                FROM baseapp_chats_message\n                WHERE room_id = OLD.room_id\n                ORDER BY created DESC\n                LIMIT 1\n            ),\n            last_message_time = (\n                SELECT created\n                FROM baseapp_chats_message\n                WHERE room_id = OLD.room_id\n                ORDER BY created DESC\n                LIMIT 1\n            )\n            WHERE id = OLD.room_id;\n            RETURN NULL;\n        ",
                            hash="185b6146e2a5bbec3fb6668180701954aebaaaa3",
                            operation="DELETE",
                            pgid="pgtrigger_update_last_message_e519b",
                            table="baseapp_chats_message",
                            when="AFTER",
                        ),
                    ),
                )
                if not swapper.is_swapped("baseapp_chats", "Message")
                else None
            ),
            (
                pgtrigger.migrations.AddTrigger(
                    model_name="messagestatus",
                    trigger=pgtrigger.compiler.Trigger(
                        name="increment_unread_count",
                        sql=pgtrigger.compiler.UpsertTriggerSql(
                            condition='WHEN (NOT NEW."is_read")',
                            func="\n            INSERT INTO baseapp_chats_unreadmessagecount (room_id, profile_id, count)\n            VALUES ((SELECT room_id FROM baseapp_chats_message WHERE id = NEW.message_id), NEW.profile_id, 1)\n            ON CONFLICT (room_id, profile_id)\n            DO UPDATE SET count = baseapp_chats_unreadmessagecount.count + 1;\n            RETURN NULL;\n        ",
                            hash="b6b105f4558ddea8d44c7872d91b78510843366b",
                            operation="INSERT",
                            pgid="pgtrigger_increment_unread_count_ec3cd",
                            table="baseapp_chats_messagestatus",
                            when="AFTER",
                        ),
                    ),
                )
                if not swapper.is_swapped("baseapp_chats", "MessageStatus")
                else None
            ),
            (
                pgtrigger.migrations.AddTrigger(
                    model_name="messagestatus",
                    trigger=pgtrigger.compiler.Trigger(
                        name="decrement_unread_count",
                        sql=pgtrigger.compiler.UpsertTriggerSql(
                            condition='WHEN (NEW."is_read" AND NOT OLD."is_read")',
                            func="\n            UPDATE baseapp_chats_unreadmessagecount\n            SET count = GREATEST(0, count - 1)\n            WHERE\n                room_id = (SELECT room_id FROM baseapp_chats_message WHERE id = NEW.message_id) AND\n                profile_id = NEW.profile_id;\n            RETURN NULL;\n        ",
                            hash="08effa83237ad55cf10b5eb25e6fc9e824799b14",
                            operation="UPDATE",
                            pgid="pgtrigger_decrement_unread_count_a1c0b",
                            table="baseapp_chats_messagestatus",
                            when="AFTER",
                        ),
                    ),
                )
                if not swapper.is_swapped("baseapp_chats", "MessageStatus")
                else None
            ),
        ],
    )
